#!/usr/bin/env python3

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2017 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.1.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

from copy import deepcopy
import gzip
import os
import sys
import tempfile
import unittest
import uuid

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
PACKAGE_DIR = os.path.dirname(TEST_DIR)
sys.path.append(PACKAGE_DIR)

from anacore.sv import SVIO, HashedSVIO


########################################################################
#
# FUNCTIONS
#
########################################################################
class TestSVIO(unittest.TestCase):
    def setUp(self):
        tmp_folder = tempfile.gettempdir()
        unique_id = str(uuid.uuid1())

        # Temporary files
        self.tmp_in_tsv = os.path.join(tmp_folder, unique_id + "_in.tsv")
        self.tmp_in_metadata_tsv = os.path.join(tmp_folder, unique_id + "_in_withMeta.tsv")
        self.tmp_in_csv = os.path.join(tmp_folder, unique_id + "_in.csv")
        self.tmp_out_tsv = os.path.join(tmp_folder, unique_id + "_out.tsv")
        self.tmp_out_dsv = os.path.join(tmp_folder, unique_id + "_out.dsv")
        self.tmp_out_dsv_gz = os.path.join(tmp_folder, unique_id + "_out.dsv.gz")

        # Data
        self.data = {
            "metadata": ["release: 17 ; date: 2021-03-01", "author: anapath"],
            "titles": ["CHROM", "POS", "ID", "REF", "ALT", "QUAL", "FILTER"],
            "rows": [
                ["20", "14370", "rs6054257", "G", "A", "29", "PASS"],
                ["20", "17330", "", "", "A", "3", "q10 low"],
                ["20", "1110696", "rs6040355", "A", "G,T", "67", "PASS"],
                ["", "1230237", "", "T", "", "47", ""],
                ["20", "1234567", "microsat1", "GTC", "G,GTCT", "50", ""],
                ["", "", "", "", "", "", ""]
            ]
        }

        # Create TSV
        tsv_content = """#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER
20\t14370\trs6054257\tG\tA\t29\tPASS
20\t17330\t   \t\tA\t3\tq10 low
20\t1110696\trs6040355\tA\tG,T\t67\tPASS
\t1230237\t\tT\t\t47\t
20\t1234567\tmicrosat1\tGTC\tG,GTCT\t50\t
\t\t\t\t\t\t"""
        with open(self.tmp_in_tsv, "w") as FH_tsv:
            FH_tsv.write(tsv_content)

        # Create metadata TSV
        metadata_tsv_content = """##release: 17 ; date: 2021-03-01
##author: anapath"""
        with open(self.tmp_in_metadata_tsv, "w") as FH_tsv:
            FH_tsv.write(metadata_tsv_content)

        # Create CSV
        csv_content = tsv_content[1:].replace("\t", ";")
        with open(self.tmp_in_csv, "w") as FH_csv:
            FH_csv.write(csv_content)

    def testIsValidTrue(self):
        # Create one valid DSV file
        rows = deepcopy(self.data["rows"])
        rows.insert(0, self.data["titles"])
        rows.extend([
            ["1", "1234567", "microsat1", "GTC", "G,GTCT", "50", ""],
            ["1", "1234567", "microsat1", "GTC", "G,GTCT", "50", ""],
            ["1", "1234567", "microsat1", "GTC", "G,GTCT", "50", ""],
            ["1", "1234567", "microsat1", "GTC", "G,GTCT", "50", ""],
            ["1", "1234567", "microsat1", "GTC", "G,GTCT", "50", ""],
            ["1", "1234567", "microsat1", "GTC", "G,GTCT", "50", ""],
            ["1", "1234567", "microsat1", "GTC", "G,GTCT", "50", ""],
            ["1", "1234567", "microsat1", "GTC", "G,GTCT", "50", ""]
        ])
        with open(self.tmp_out_dsv, "w") as FH_out:
            for curr_row in rows:
                FH_out.write(".".join(curr_row) + "\n")
        # Check isValid
        self.assertEqual(True, SVIO.isValid(self.tmp_out_dsv, "."))
        # Check if fie pointer is ok in reopen
        with open(self.tmp_out_dsv) as FH_in:
            for row_idx, readed_row in enumerate(FH_in):
                expected_row = ".".join(rows[row_idx]) + "\n"
                self.assertEqual(expected_row, readed_row)

    def testIsValidWithMetadata(self):
        # Create one valid DSV file
        content = """@title: test
#CHROM.POS.ID.REF.ALT.QUAL.FILTER
20.14370.rs6054257.G.A.29.PASS
20.17330.   ..A.3.q10 low

20.1110696.rs6040355.A.G,T.67.PASS
.1230237..T..47.
20.1234567.microsat1.GTC.G,GTCT.50.
......"""
        with open(self.tmp_out_dsv, "w") as FH_out:
            FH_out.write(content)
        # Check isValid
        self.assertTrue(not SVIO.isValid(self.tmp_out_dsv, ".", "@"))
        # Create one valid DSV file
        content = """@title: test
#CHROM.POS.ID.REF.ALT.QUAL.FILTER
20.14370.rs6054257.G.A.29.PASS
20.17330.   ..A.3.q10 low
20.1110696.rs6040355.A.G,T.67.PASS
.1230237..T..47.
20.1234567.microsat1.GTC.G,GTCT.50.
......"""
        with open(self.tmp_out_dsv, "w") as FH_out:
            FH_out.write(content)
        # Check isValid
        self.assertTrue(SVIO.isValid(self.tmp_out_dsv, ".", "@"))
        # Check if file pointer is ok in reopen
        observed_rows = []
        with open(self.tmp_out_dsv) as FH_in:
            for row_idx, readed_row in enumerate(FH_in):
                observed_rows.append(readed_row)
        self.assertEqual(content, "".join(observed_rows))

    def testIsValidFalse(self):
        # Create one invalid TSV file
        rows = deepcopy(self.data["rows"])
        rows.insert(0, self.data["titles"])
        rows.extend([
            ["1", "1234567", "microsat1", "GTC", "G,GTCT", "50", ""],
            ["1", "1234567", "microsat1", "GTC", "G,GTCT", "50", ""],
            ["1", "microsat1", "GTC", "G,GTCT", "50", ""],
            ["1", "1234567", "microsat1", "GTC", "G,GTCT", "50", ""],
            ["1", "1234567", "microsat1", "GTC", "G,GTCT", "50", ""],
            ["1", "1234567", "microsat1", "GTC", "G,GTCT", "50", ""],
            ["1", "1234567", "microsat1", "GTC", "G,GTCT", "50", ""],
            ["1", "1234567", "microsat1", "GTC", "G,GTCT", "50", ""]
        ])
        with open(self.tmp_out_dsv, "w") as FH_out:
            for curr_row in rows:
                FH_out.write(".".join(curr_row) + "\n")
        # Check isValid
        self.assertEqual(False, SVIO.isValid(self.tmp_out_dsv, "."))
        # Check if file pointer is ok in reopen
        with open(self.tmp_out_dsv) as FH_in:
            for row_idx, readed_row in enumerate(FH_in):
                expected_row = ".".join(rows[row_idx]) + "\n"
                self.assertEqual(expected_row, readed_row)

    def testIterCSV(self):
        with SVIO(self.tmp_in_csv, separator=";", title_starter=None) as FH_in:
            # Header
            self.assertEqual(FH_in.titles, self.data["titles"])
            # Records
            for record_idx, readed_records in enumerate(FH_in):
                self.assertEqual(self.data["rows"][record_idx], readed_records)

    def testIterTSV(self):
        with SVIO(self.tmp_in_tsv, title_starter="#") as FH_in:
            # Header
            self.assertEqual(FH_in.titles, self.data["titles"])
            # Records
            for record_idx, readed_records in enumerate(FH_in):
                self.assertEqual(self.data["rows"][record_idx], readed_records)

    def testWriteDSV(self):
        # Write file
        with SVIO(self.tmp_out_dsv, "w", separator=".", title_starter="#") as FH_out:
            FH_out.titles = self.data["titles"]
            for row in self.data["rows"]:
                FH_out.write(row)
        # Assert result
        with open(self.tmp_out_dsv) as FH_in:
            for row_idx, readed_row in enumerate(FH_in):
                expected_row = ""
                if row_idx == 0:
                    expected_row = "#" + ".".join(self.data["titles"]) + "\n"
                else:
                    expected_row = ".".join(self.data["rows"][row_idx - 1]) + "\n"
                self.assertEqual(expected_row, readed_row)

    def testAppendDSV(self):
        split_limit = 3  # Numbers of row in first write
        # 1st write step
        with SVIO(self.tmp_out_dsv, "a", separator=".", title_starter="#") as FH_out:
            FH_out.titles = self.data["titles"]
            for row in self.data["rows"][:split_limit]:
                FH_out.write(row)
        # Assert result
        with open(self.tmp_out_dsv) as FH_in:
            nb_rows = -1
            for row_idx, readed_row in enumerate(FH_in):
                nb_rows += 1
                expected_row = ""
                if row_idx == 0:
                    expected_row = "#" + ".".join(self.data["titles"]) + "\n"
                else:
                    expected_row = ".".join(self.data["rows"][row_idx - 1]) + "\n"
                self.assertEqual(expected_row, readed_row)
            self.assertEqual(split_limit, nb_rows)
        # 2nd write step
        with SVIO(self.tmp_out_dsv, "a", separator=".", title_starter="#") as FH_out:
            self.assertEqual(self.data["titles"], FH_out.titles)  # Assert titles retrieval
            self.assertEqual(split_limit + 1, FH_out.current_line_nb)
            for row in self.data["rows"][split_limit:]:
                FH_out.write(row)
        # Assert result
        with open(self.tmp_out_dsv) as FH_in:
            nb_rows = -1
            for row_idx, readed_row in enumerate(FH_in):
                nb_rows += 1
                expected_row = ""
                if row_idx == 0:
                    expected_row = "#" + ".".join(self.data["titles"]) + "\n"
                else:
                    expected_row = ".".join(self.data["rows"][row_idx - 1]) + "\n"
                self.assertEqual(expected_row, readed_row)
            self.assertEqual(len(self.data["rows"]), nb_rows)

    def testWriteHeader(self):
        # With titles and metadata
        with SVIO(self.tmp_out_tsv, "w") as writer:
            writer.titles = self.data["titles"]
            writer.metadata = self.data["metadata"]
            writer.writeHeader()
            writer.write(self.data["rows"][0])
        with SVIO(self.tmp_out_tsv, "r") as reader:
            self.assertEqual(reader.titles, self.data["titles"])
            self.assertEqual(reader.metadata, self.data["metadata"])
            self.assertEqual(reader.read(), [self.data["rows"][0]])
        # Without metadata
        with SVIO(self.tmp_out_tsv, "w") as writer:
            writer.titles = self.data["titles"]
            writer.writeHeader()
            writer.write(self.data["rows"][0])
        with SVIO(self.tmp_out_tsv, "r") as reader:
            self.assertEqual(reader.titles, self.data["titles"])
            self.assertEqual(reader.metadata, [])
            self.assertEqual(reader.read(), [self.data["rows"][0]])
        # Without titles
        with SVIO(self.tmp_out_tsv, "w", has_title=False) as writer:
            writer.metadata = self.data["metadata"]
            writer.writeHeader()
            writer.write(self.data["rows"][0])
        with SVIO(self.tmp_out_tsv, "r", has_title=False) as reader:
            self.assertEqual(reader.titles, None)
            self.assertEqual(reader.metadata, self.data["metadata"])
            self.assertEqual(reader.read(), [self.data["rows"][0]])
        # Without titles and metadata
        with SVIO(self.tmp_out_tsv, "w", has_title=False) as writer:
            writer.writeHeader()
            writer.write(self.data["rows"][0])
        with SVIO(self.tmp_out_tsv, "r", has_title=False) as reader:
            self.assertEqual(reader.titles, None)
            self.assertEqual(reader.metadata, [])
            self.assertEqual(reader.read(), [self.data["rows"][0]])
        # Only header with titles and metadata
        with SVIO(self.tmp_out_tsv, "w") as writer:
            writer.titles = self.data["titles"]
            writer.metadata = self.data["metadata"]
            writer.writeHeader()
        with SVIO(self.tmp_out_tsv, "r") as reader:
            self.assertEqual(reader.titles, self.data["titles"])
            self.assertEqual(reader.metadata, self.data["metadata"])
        # Only header without metadata
        with SVIO(self.tmp_out_tsv, "w") as writer:
            writer.titles = self.data["titles"]
            writer.writeHeader()
        with SVIO(self.tmp_out_tsv, "r") as reader:
            self.assertEqual(reader.titles, self.data["titles"])
            self.assertEqual(reader.metadata, [])
        # Only header without titles
        with SVIO(self.tmp_out_tsv, "w", has_title=False) as writer:
            writer.metadata = self.data["metadata"]
            writer.writeHeader()
        with SVIO(self.tmp_out_tsv, "r", has_title=False) as reader:
            self.assertEqual(reader.titles, None)
            self.assertEqual(reader.metadata, self.data["metadata"])
        # Empty without titles and metadata
        with SVIO(self.tmp_out_tsv, "w", has_title=False) as writer:
            writer.writeHeader()
        with SVIO(self.tmp_out_tsv, "r", has_title=False) as reader:
            self.assertEqual(reader.titles, None)
            self.assertEqual(reader.metadata, [])

    def testAppendEmptyDSV(self):
        # Write file
        with SVIO(self.tmp_out_dsv, "a", separator=".", title_starter="#") as FH_out:
            self.assertEqual(0, FH_out.current_line_nb)
            FH_out.titles = self.data["titles"]
            for row in self.data["rows"]:
                FH_out.write(row)
        # Assert result
        with open(self.tmp_out_dsv) as FH_in:
            for row_idx, readed_row in enumerate(FH_in):
                expected_row = ""
                if row_idx == 0:
                    expected_row = "#" + ".".join(self.data["titles"]) + "\n"
                else:
                    expected_row = ".".join(self.data["rows"][row_idx - 1]) + "\n"
                self.assertEqual(expected_row, readed_row)

    def testAppendCompressedEmptyDSV(self):
        # Cretes empty file
        with gzip.open(self.tmp_out_dsv_gz, "wt") as FH_out:
            pass
        # Write file
        with SVIO(self.tmp_out_dsv_gz, "a", separator=".", title_starter="#") as FH_out:
            FH_out.titles = self.data["titles"]
            for row in self.data["rows"]:
                FH_out.write(row)
        # Assert result
        nb_rows = -1
        with gzip.open(self.tmp_out_dsv_gz, "rt") as FH_in:
            for row_idx, readed_row in enumerate(FH_in):
                nb_rows += 1
                expected_row = ""
                if row_idx == 0:
                    expected_row = "#" + ".".join(self.data["titles"]) + "\n"
                else:
                    expected_row = ".".join(self.data["rows"][row_idx - 1]) + "\n"
                self.assertEqual(expected_row, readed_row)
        self.assertEqual(len(self.data["rows"]), nb_rows)

    def testAppendCompressedDSV(self):
        split_limit = 3  # Numbers of row in first write
        # 1st write step
        with SVIO(self.tmp_out_dsv_gz, "a", separator=".", title_starter="#") as FH_out:
            FH_out.titles = self.data["titles"]
            for row in self.data["rows"][:split_limit]:
                FH_out.write(row)
        # Assert result
        nb_rows = -1
        with gzip.open(self.tmp_out_dsv_gz, "rt") as FH_in:
            for row_idx, readed_row in enumerate(FH_in):
                nb_rows += 1
                expected_row = ""
                if row_idx == 0:
                    expected_row = "#" + ".".join(self.data["titles"]) + "\n"
                else:
                    expected_row = ".".join(self.data["rows"][row_idx - 1]) + "\n"
                self.assertEqual(expected_row, readed_row)
        self.assertEqual(split_limit, nb_rows)
        # 2st write step
        with SVIO(self.tmp_out_dsv_gz, "a", separator=".", title_starter="#") as FH_out:
            self.assertEqual(self.data["titles"], FH_out.titles)  # Assert titles retrieval
            for row in self.data["rows"][split_limit:]:
                FH_out.write(row)
        # Assert result
        nb_rows = -1
        with gzip.open(self.tmp_out_dsv_gz, "rt") as FH_in:
            for row_idx, readed_row in enumerate(FH_in):
                nb_rows += 1
                expected_row = ""
                if row_idx == 0:
                    expected_row = "#" + ".".join(self.data["titles"]) + "\n"
                else:
                    expected_row = ".".join(self.data["rows"][row_idx - 1]) + "\n"
                self.assertEqual(expected_row, readed_row)
        self.assertEqual(len(self.data["rows"]), nb_rows)

    def tearDown(self):
        # Clean temporary files
        for curr_file in [self.tmp_in_tsv, self.tmp_in_metadata_tsv, self.tmp_in_csv, self.tmp_out_tsv, self.tmp_out_dsv, self.tmp_out_dsv_gz]:
            if os.path.exists(curr_file):
                os.remove(curr_file)


########################################################################
#
# MAIN
#
########################################################################
if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2018 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.2.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

import os
import sys
import uuid
import tempfile
import unittest

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
PACKAGE_DIR = os.path.dirname(TEST_DIR)
sys.path.append(PACKAGE_DIR)

from anacore.gtf import GTFIO, loadModel
from anacore.region import RegionTree
from anacore.genomicRegion import Gene, Transcript, Protein, Exon, CDS


########################################################################
#
# FUNCTIONS
#
########################################################################
class TestGTFIO(unittest.TestCase):
    def setUp(self):
        tmp_folder = tempfile.gettempdir()
        unique_id = str(uuid.uuid1())

        # Temporary files
        self.tmp_ensembl_in_gtf = os.path.join(tmp_folder, unique_id + "_ensembl_in.gtf")
        self.tmp_ensembl_out_gtf = os.path.join(tmp_folder, unique_id + "_ensembl_out.gtf")
        self.tmp_ncbi_in_gtf = os.path.join(tmp_folder, unique_id + "_ncbi_in.gtf")
        self.tmp_ncbi_out_gtf = os.path.join(tmp_folder, unique_id + "_ncbi_out.gtf")

        # Ensembl GTF
        self.ensembl_expected = [
            Gene(54770583, 54771134, "+", "6", "KRASP1", {"id": "ENSG00000220635"}, None, [
                Transcript(54770583, 54771134, "+", "6", "KRASP1-201", {"id": "ENST00000407852"}, None, [
                    Exon(54770583, 54771134, "+", "6", "ENST00000407852_e1")
                ])
            ]),
            Gene(25204789, 25250936, "-", "12", "KRAS", {"id": "ENSG00000133703"}, None, [
                Transcript(25204789, 25250931, "-", "12", "KRAS-202", {"id": "ENST00000311936"}, None,
                    [
                        Exon(25250751, 25250931, "-", "12", "ENST00000311936_e1"),
                        Exon(25245274, 25245395, "-", "12", "ENST00000311936_e2"),
                        Exon(25227234, 25227412, "-", "12", "ENST00000311936_e3"),
                        Exon(25225614, 25225773, "-", "12", "ENST00000311936_e5"),
                        Exon(25204789, 25209911, "-", "12", "ENST00000311936_e6")
                    ], [
                        Protein(25209798, 25245384, "-", "12", None, None, None,
                            [
                                CDS(25245274, 25245384, "-", "12", ""),
                                CDS(25227234, 25227412, "-", "12", ""),
                                CDS(25225614, 25225773, "-", "12", ""),
                                CDS(25209798, 25209911, "-", "12", "")
                            ]
                        )
                    ]
                ),
                Transcript(25209168, 25250936, "-", "12", "KRAS-204", {"id": "ENST00000557334"}, None,
                    [
                        Exon(25250751, 25250936, "-", "12", "ENST00000557334_e1"),
                        Exon(25245274, 25245395, "-", "12", "ENST00000557334_e2"),
                        Exon(25209168, 25209911, "-", "12", "ENST00000557334_e3")
                    ], [
                        Protein(25209798, 25245384, "-", "12", None, None, None,
                            [
                                CDS(25245274, 25245384, "-", "12", ""),
                                CDS(25209798, 25209911, "-", "12", "")
                            ]
                        )
                    ]
                ),
                Transcript(25209431, 25250803, "-", "12", "KRAS-201", {"id": "ENST00000256078"}, None,
                    [
                        Exon(25250751, 25250803, "-", "12", "ENST00000256078_e1"),
                        Exon(25245274, 25245395, "-", "12", "ENST00000256078_e2"),
                        Exon(25227234, 25227412, "-", "12", "ENST00000256078_e3"),
                        Exon(25225614, 25225773, "-", "12", "ENST00000256078_e4"),
                        Exon(25215437, 25215560, "-", "12", "ENST00000256078_e5"),
                        Exon(25209431, 25209911, "-", "12", "ENST00000256078_e6")
                    ], [
                        Protein(25215444, 25245384, "-", "12", None, None, None,
                            [
                                CDS(25245274, 25245384, "-", "12", ""),
                                CDS(25227234, 25227412, "-", "12", ""),
                                CDS(25225614, 25225773, "-", "12", ""),
                                CDS(25215444, 25215560, "-", "12", "")
                            ]
                        )
                    ]
                ),
                Transcript(25233819, 25250929, "-", "12", "KRAS-203", {"id": "ENST00000556131"}, None,
                    [
                        Exon(25250764, 25250929, "-", "12", "ENST00000556131_e1"),
                        Exon(25245274, 25245395, "-", "12", "ENST00000556131_e2"),
                        Exon(25233819, 25235226, "-", "12", "ENST00000556131_e3")
                    ], [
                        Protein(25235209, 25245384, "-", "12", None, None, None,
                            [
                                CDS(25245274, 25245384, "-", "12", ""),
                                CDS(25235209, 25235226, "-", "12", "")
                            ]
                        )
                    ]
                )
            ])
        ]
        with open(self.tmp_ensembl_in_gtf, "w") as FH_gtf:
            FH_gtf.write("""#!genome-build GRCh38.p12
#!genome-version GRCh38
#!genome-date 2013-12
#!genome-build-accession NCBI:GCA_000001405.27
#!genebuild-last-updated 2018-07
6	havana	gene	54770583	54771134	.	+	.	gene_id \"ENSG00000220635\"; gene_version \"2\"; gene_name \"KRASP1\"; gene_source \"havana\"; gene_biotype \"processed_pseudogene\";
6	havana	transcript	54770583	54771134	.	+	.	gene_id \"ENSG00000220635\"; gene_version \"2\"; transcript_id \"ENST00000407852\"; transcript_version \"2\"; gene_name \"KRASP1\"; gene_source \"havana\"; gene_biotype \"processed_pseudogene\"; transcript_name \"KRASP1-201\"; transcript_source \"havana\"; transcript_biotype \"processed_pseudogene\"; tag \"basic\"; transcript_support_level \"NA\";
6	havana	exon	54770583	54771134	.	+	.	gene_id \"ENSG00000220635\"; gene_version \"2\"; transcript_id \"ENST00000407852\"; transcript_version \"2\"; exon_number \"1\"; gene_name \"KRASP1\"; gene_source \"havana\"; gene_biotype \"processed_pseudogene\"; transcript_name \"KRASP1-201\"; transcript_source \"havana\"; transcript_biotype \"processed_pseudogene\"; exon_id \"ENSE00001550689\"; exon_version \"2\"; tag \"basic\"; transcript_support_level \"NA\";
12	ensembl_havana	gene	25204789	25250936	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\";
12	ensembl_havana	transcript	25204789	25250931	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000311936\"; transcript_version \"7\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-202\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8702\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	exon	25250751	25250931	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000311936\"; transcript_version \"7\"; exon_number \"1\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-202\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8702\"; exon_id \"ENSE00001189804\"; exon_version \"4\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	exon	25245274	25245395	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000311936\"; transcript_version \"7\"; exon_number \"2\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-202\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8702\"; exon_id \"ENSE00000936617\"; exon_version \"1\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	CDS	25245274	25245384	.	-	0	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000311936\"; transcript_version \"7\"; exon_number \"2\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-202\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8702\"; protein_id \"ENSP00000308495\"; protein_version \"3\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	start_codon	25245382	25245384	.	-	0	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000311936\"; transcript_version \"7\"; exon_number \"2\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-202\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8702\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	exon	25227234	25227412	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000311936\"; transcript_version \"7\"; exon_number \"3\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-202\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8702\"; exon_id \"ENSE00001719809\"; exon_version \"1\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	CDS	25227234	25227412	.	-	0	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000311936\"; transcript_version \"7\"; exon_number \"3\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-202\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8702\"; protein_id \"ENSP00000308495\"; protein_version \"3\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	exon	25225614	25225773	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000311936\"; transcript_version \"7\"; exon_number \"4\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-202\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8702\"; exon_id \"ENSE00001644818\"; exon_version \"1\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	CDS	25225614	25225773	.	-	1	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000311936\"; transcript_version \"7\"; exon_number \"4\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-202\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8702\"; protein_id \"ENSP00000308495\"; protein_version \"3\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	exon	25204789	25209911	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000311936\"; transcript_version \"7\"; exon_number \"5\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-202\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8702\"; exon_id \"ENSE00002456976\"; exon_version \"1\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	CDS	25209798	25209911	.	-	0	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000311936\"; transcript_version \"7\"; exon_number \"5\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-202\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8702\"; protein_id \"ENSP00000308495\"; protein_version \"3\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	stop_codon	25209795	25209797	.	-	0	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000311936\"; transcript_version \"7\"; exon_number \"5\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-202\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8702\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	five_prime_utr	25250751	25250931	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000311936\"; transcript_version \"7\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-202\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8702\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	five_prime_utr	25245385	25245395	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000311936\"; transcript_version \"7\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-202\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8702\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	three_prime_utr	25204789	25209794	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000311936\"; transcript_version \"7\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-202\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8702\"; tag \"basic\"; transcript_support_level \"1\";
12	havana	transcript	25209168	25250936	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000557334\"; transcript_version \"5\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-204\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; tag \"basic\"; transcript_support_level \"5\";
12	havana	exon	25250751	25250936	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000557334\"; transcript_version \"5\"; exon_number \"1\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-204\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; exon_id \"ENSE00002446502\"; exon_version \"1\"; tag \"basic\"; transcript_support_level \"5\";
12	havana	exon	25245274	25245395	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000557334\"; transcript_version \"5\"; exon_number \"2\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-204\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; exon_id \"ENSE00000936617\"; exon_version \"1\"; tag \"basic\"; transcript_support_level \"5\";
12	havana	CDS	25245274	25245384	.	-	0	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000557334\"; transcript_version \"5\"; exon_number \"2\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-204\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; protein_id \"ENSP00000452512\"; protein_version \"1\"; tag \"basic\"; transcript_support_level \"5\";
12	havana	start_codon	25245382	25245384	.	-	0	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000557334\"; transcript_version \"5\"; exon_number \"2\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-204\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; tag \"basic\"; transcript_support_level \"5\";
12	havana	exon	25209168	25209911	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000557334\"; transcript_version \"5\"; exon_number \"3\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-204\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; exon_id \"ENSE00002464674\"; exon_version \"1\"; tag \"basic\"; transcript_support_level \"5\";
12	havana	CDS	25209798	25209911	.	-	0	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000557334\"; transcript_version \"5\"; exon_number \"3\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-204\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; protein_id \"ENSP00000452512\"; protein_version \"1\"; tag \"basic\"; transcript_support_level \"5\";
12	havana	stop_codon	25209795	25209797	.	-	0	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000557334\"; transcript_version \"5\"; exon_number \"3\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-204\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; tag \"basic\"; transcript_support_level \"5\";
12	havana	five_prime_utr	25250751	25250936	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000557334\"; transcript_version \"5\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-204\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; tag \"basic\"; transcript_support_level \"5\";
12	havana	five_prime_utr	25245385	25245395	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000557334\"; transcript_version \"5\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-204\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; tag \"basic\"; transcript_support_level \"5\";
12	havana	three_prime_utr	25209168	25209794	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000557334\"; transcript_version \"5\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-204\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; tag \"basic\"; transcript_support_level \"5\";
12	ensembl_havana	transcript	25209431	25250803	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000256078\"; transcript_version \"8\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-201\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8703\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	exon	25250751	25250803	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000256078\"; transcript_version \"8\"; exon_number \"1\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-201\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8703\"; exon_id \"ENSE00002513959\"; exon_version \"1\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	exon	25245274	25245395	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000256078\"; transcript_version \"8\"; exon_number \"2\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-201\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8703\"; exon_id \"ENSE00000936617\"; exon_version \"1\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	CDS	25245274	25245384	.	-	0	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000256078\"; transcript_version \"8\"; exon_number \"2\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-201\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8703\"; protein_id \"ENSP00000256078\"; protein_version \"4\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	start_codon	25245382	25245384	.	-	0	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000256078\"; transcript_version \"8\"; exon_number \"2\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-201\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8703\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	exon	25227234	25227412	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000256078\"; transcript_version \"8\"; exon_number \"3\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-201\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8703\"; exon_id \"ENSE00001719809\"; exon_version \"1\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	CDS	25227234	25227412	.	-	0	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000256078\"; transcript_version \"8\"; exon_number \"3\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-201\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8703\"; protein_id \"ENSP00000256078\"; protein_version \"4\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	exon	25225614	25225773	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000256078\"; transcript_version \"8\"; exon_number \"4\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-201\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8703\"; exon_id \"ENSE00001644818\"; exon_version \"1\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	CDS	25225614	25225773	.	-	1	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000256078\"; transcript_version \"8\"; exon_number \"4\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-201\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8703\"; protein_id \"ENSP00000256078\"; protein_version \"4\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	exon	25215437	25215560	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000256078\"; transcript_version \"8\"; exon_number \"5\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-201\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8703\"; exon_id \"ENSE00001189807\"; exon_version \"5\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	CDS	25215444	25215560	.	-	0	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000256078\"; transcript_version \"8\"; exon_number \"5\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-201\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8703\"; protein_id \"ENSP00000256078\"; protein_version \"4\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	stop_codon	25215441	25215443	.	-	0	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000256078\"; transcript_version \"8\"; exon_number \"5\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-201\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8703\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	exon	25209431	25209911	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000256078\"; transcript_version \"8\"; exon_number \"6\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-201\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8703\"; exon_id \"ENSE00002477035\"; exon_version \"1\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	five_prime_utr	25250751	25250803	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000256078\"; transcript_version \"8\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-201\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8703\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	five_prime_utr	25245385	25245395	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000256078\"; transcript_version \"8\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-201\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8703\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	three_prime_utr	25215437	25215440	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000256078\"; transcript_version \"8\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-201\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8703\"; tag \"basic\"; transcript_support_level \"1\";
12	ensembl_havana	three_prime_utr	25209431	25209911	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000256078\"; transcript_version \"8\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-201\"; transcript_source \"ensembl_havana\"; transcript_biotype \"protein_coding\"; tag \"CCDS\"; ccds_id \"CCDS8703\"; tag \"basic\"; transcript_support_level \"1\";
12	havana	transcript	25233819	25250929	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000556131\"; transcript_version \"1\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-203\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; tag \"basic\"; transcript_support_level \"1\";
12	havana	exon	25250764	25250929	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000556131\"; transcript_version \"1\"; exon_number \"1\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-203\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; exon_id \"ENSE00002530521\"; exon_version \"1\"; tag \"basic\"; transcript_support_level \"1\";
12	havana	exon	25245274	25245395	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000556131\"; transcript_version \"1\"; exon_number \"2\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-203\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; exon_id \"ENSE00000936617\"; exon_version \"1\"; tag \"basic\"; transcript_support_level \"1\";
12	havana	CDS	25245274	25245384	.	-	0	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000556131\"; transcript_version \"1\"; exon_number \"2\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-203\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; protein_id \"ENSP00000451856\"; protein_version \"1\"; tag \"basic\"; transcript_support_level \"1\";
12	havana	start_codon	25245382	25245384	.	-	0	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000556131\"; transcript_version \"1\"; exon_number \"2\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-203\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; tag \"basic\"; transcript_support_level \"1\";
12	havana	exon	25233819	25235226	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000556131\"; transcript_version \"1\"; exon_number \"3\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-203\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; exon_id \"ENSE00002478081\"; exon_version \"1\"; tag \"basic\"; transcript_support_level \"1\";
12	havana	CDS	25235209	25235226	.	-	0	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000556131\"; transcript_version \"1\"; exon_number \"3\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-203\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; protein_id \"ENSP00000451856\"; protein_version \"1\"; tag \"basic\"; transcript_support_level \"1\";
12	havana	stop_codon	25235206	25235208	.	-	0	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000556131\"; transcript_version \"1\"; exon_number \"3\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-203\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; tag \"basic\"; transcript_support_level \"1\";
12	havana	five_prime_utr	25250764	25250929	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000556131\"; transcript_version \"1\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-203\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; tag \"basic\"; transcript_support_level \"1\";
12	havana	five_prime_utr	25245385	25245395	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000556131\"; transcript_version \"1\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-203\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; tag \"basic; other\"; transcript_support_level \"1\";
12	havana	three_prime_utr	25233819	25235205	.	-	.	gene_id \"ENSG00000133703\"; gene_version \"11\"; transcript_id \"ENST00000556131\"; transcript_version \"1\"; gene_name \"KRAS\"; gene_source \"ensembl_havana\"; gene_biotype \"protein_coding\"; transcript_name \"KRAS-203\"; transcript_source \"havana\"; transcript_biotype \"protein_coding\"; tag \"basic;\"; transcript_support_level \"1\";""")


        # NCBI GTF
        self.ncbi_expected = [
            Gene(54635272, 54640529, "+", "6", "KRASP1", {"id": "3844"}, None, [
                Transcript(54635272, 54640529, "+", "6", "gene14201", {"id": "gene14201"}, None, [
                    Exon(54635272, 54640529, "+", "6", "gene14201_e1")
                ])
            ]),
            Gene(25357723, 25403865, "-", "12", "KRAS", {"id": "3845"}, None, [
                Transcript(25357723, 25403865, "-", "12", "rna36549", {"id": "rna36549"}, None,
                    [
                        Exon(25357723, 25362845, "-", "12", "rna36549_e1"),
                        Exon(25378548, 25378707, "-", "12", "rna36549_e2"),
                        Exon(25380168, 25380346, "-", "12", "rna36549_e3"),
                        Exon(25398208, 25398329, "-", "12", "rna36549_e5"),
                        Exon(25403685, 25403865, "-", "12", "rna36549_e6")
                    ], [
                        Protein(25362729, 25398318, "-", "12", None, None, None,
                            [
                                CDS(25362729, 25362845, "-", "12", ""),
                                CDS(25378548, 25378707, "-", "12", ""),
                                CDS(25380168, 25380346, "-", "12", ""),
                                CDS(25398208, 25398318, "-", "12", "")
                            ]
                        )
                    ]
                ),
                Transcript(25357723, 25403865, "-", "12", "rna36550", {"id": "rna36550"}, None,
                    [
                        Exon(25357723, 25362845, "-", "12", "rna36550_e1"),
                        Exon(25368371, 25368494, "-", "12", "rna36550_e2"),
                        Exon(25378548, 25378707, "-", "12", "rna36550_e3"),
                        Exon(25380168, 25380346, "-", "12", "rna36550_e4"),
                        Exon(25398208, 25398329, "-", "12", "rna36550_e5"),
                        Exon(25403685, 25403865, "-", "12", "rna36550_e6")
                    ], [
                        Protein(25368375, 25398318, "-", "12", None, None, None,
                            [
                                CDS(25368375, 25368494, "-", "12", ""),
                                CDS(25378548, 25378707, "-", "12", ""),
                                CDS(25380168, 25380346, "-", "12", ""),
                                CDS(25398208, 25398318, "-", "12", ""),
                            ]
                        )
                    ]
                )
            ]),
            Gene(2527306, 2529079, "+", "X", "CD99P1", {"id": "401577"}, None, [
                Transcript(2527306, 2529079, "+", "X", "rna58916", {"id": "rna58916"}, None, [
                    Exon(2527306, 2527522, "+", "X", "rna58916_e1"),
                    Exon(2529037, 2529079, "+", "X", "rna58916_e2")
                ])
            ]),
            Gene(2477306, 2479079, "+", "Y", "CD99P1", {"id": "401577"}, None, [
                Transcript(2477306, 2479079, "+", "Y", "rna61353", {"id": "rna61353"}, None, [
                    Exon(2477306, 2477522, "+", "Y", "rna61353_e1"),
                    Exon(2479037, 2479079, "+", "Y", "rna61353_e2")
                ])
            ])
        ]
        with open(self.tmp_ncbi_in_gtf, "w") as FH_gtf:
            FH_gtf.write("""6	Curated Genomic	exon	54635272	54640529	.	+	.	transcript_id \"gene14201\"; gene_id \"3844\"; gene_name \"KRASP1\";
12	BestRefSeq	exon	25357723	25362845	.	-	.	transcript_id \"rna36549\"; gene_id \"3845\"; gene_name \"KRAS\";
12	BestRefSeq	exon	25378548	25378707	.	-	.	transcript_id \"rna36549\"; gene_id \"3845\"; gene_name \"KRAS\";
12	BestRefSeq	exon	25380168	25380346	.	-	.	transcript_id \"rna36549\"; gene_id \"3845\"; gene_name \"KRAS\";
12	BestRefSeq	exon	25398208	25398329	.	-	.	transcript_id \"rna36549\"; gene_id \"3845\"; gene_name \"KRAS\";
12	BestRefSeq	exon	25403685	25403865	.	-	.	transcript_id \"rna36549\"; gene_id \"3845\"; gene_name \"KRAS\";
12	BestRefSeq	CDS	25362729	25362845	.	-	0	transcript_id \"rna36549\"; gene_id \"3845\"; gene_name \"KRAS\";
12	BestRefSeq	CDS	25378548	25378707	.	-	1	transcript_id \"rna36549\"; gene_id \"3845\"; gene_name \"KRAS\";
12	BestRefSeq	CDS	25380168	25380346	.	-	0	transcript_id \"rna36549\"; gene_id \"3845\"; gene_name \"KRAS\";
12	BestRefSeq	CDS	25398208	25398318	.	-	0	transcript_id \"rna36549\"; gene_id \"3845\"; gene_name \"KRAS\";
12	BestRefSeq	exon	25357723	25362845	.	-	.	transcript_id \"rna36550\"; gene_id \"3845\"; gene_name \"KRAS\";
12	BestRefSeq	exon	25368371	25368494	.	-	.	transcript_id \"rna36550\"; gene_id \"3845\"; gene_name \"KRAS\";
12	BestRefSeq	exon	25378548	25378707	.	-	.	transcript_id \"rna36550\"; gene_id \"3845\"; gene_name \"KRAS\";
12	BestRefSeq	exon	25380168	25380346	.	-	.	transcript_id \"rna36550\"; gene_id \"3845\"; gene_name \"KRAS\";
12	BestRefSeq	exon	25398208	25398329	.	-	.	transcript_id \"rna36550\"; gene_id \"3845\"; gene_name \"KRAS\";
12	BestRefSeq	exon	25403685	25403865	.	-	.	transcript_id \"rna36550\"; gene_id \"3845\"; gene_name \"KRAS\";
12	BestRefSeq	CDS	25368375	25368494	.	-	0	transcript_id \"rna36550\"; gene_id \"3845\"; gene_name \"KRAS\";
12	BestRefSeq	CDS	25378548	25378707	.	-	1	transcript_id \"rna36550\"; gene_id \"3845\"; gene_name \"KRAS\";
12	BestRefSeq	CDS	25380168	25380346	.	-	0	transcript_id \"rna36550\"; gene_id \"3845\"; gene_name \"KRAS\";
12	BestRefSeq	CDS	25398208	25398318	.	-	0	transcript_id \"rna36550\"; gene_id \"3845\"; gene_name \"KRAS\";
X	BestRefSeq	exon	2527306	2527522	.	+	.	transcript_id \"rna58916\"; gene_id \"401577\"; gene_name \"CD99P1\";
X	BestRefSeq	exon	2529037	2529079	.	+	.	transcript_id \"rna58916\"; gene_id \"401577\"; gene_name \"CD99P1\";
Y	BestRefSeq	exon	2477306	2477522	.	+	.	transcript_id \"rna61353\"; gene_id \"401577\"; gene_name \"CD99P1\";
Y	BestRefSeq	exon	2479037	2479079	.	+	.	transcript_id \"rna61353\"; gene_id \"401577\"; gene_name \"CD99P1\";""")


    def testLoadModelEnsembl(self):
        ensembl_genes = loadModel(self.tmp_ensembl_in_gtf, "genes")
        self.assertEqual(
            toBracketTree(self.ensembl_expected),
            toBracketTree(ensembl_genes)
        )

    def testLoadModelNCBI(self):
        ncbi_genes = loadModel(self.tmp_ncbi_in_gtf, "genes")
        self.assertEqual(
            toBracketTree(self.ncbi_expected),
            toBracketTree(ncbi_genes)
        )

    def testWriteEnsembl(self):
        # Read expected
        in_records = None
        with GTFIO(self.tmp_ensembl_in_gtf) as FH_in:
            in_records = FH_in.read()
        # Write
        with GTFIO(self.tmp_ensembl_out_gtf, "w") as FH_out:
            for record in in_records:
                FH_out.write(record)
        # Read writted
        out_records = None
        with GTFIO(self.tmp_ensembl_in_gtf) as FH_in:
            out_records = FH_in.read()
        # Compare expected and observed
        for exp_rec, obs_rec in zip(in_records, out_records):
            self.assertEqual(
                recordToStr(exp_rec),
                recordToStr(obs_rec)
            )

    def testWriteNCBI(self):
        # Read expected
        in_records = None
        with GTFIO(self.tmp_ncbi_in_gtf) as FH_in:
            in_records = FH_in.read()
        # Write
        with GTFIO(self.tmp_ncbi_out_gtf, "w") as FH_out:
            for record in in_records:
                FH_out.write(record)
        # Read writted
        out_records = None
        with GTFIO(self.tmp_ncbi_in_gtf) as FH_in:
            out_records = FH_in.read()
        # Compare expected and observed
        for exp_rec, obs_rec in zip(in_records, out_records):
            self.assertEqual(
                recordToStr(exp_rec),
                recordToStr(obs_rec)
            )

    def tearDown(self):
        # Clean temporary files
        for curr_file in [self.tmp_ensembl_in_gtf, self.tmp_ensembl_out_gtf, self.tmp_ncbi_in_gtf, self.tmp_ncbi_out_gtf]:
            if os.path.exists(curr_file):
                os.remove(curr_file)


def recordToStr(record):
    attributes = []
    for key, val in sorted(record.annot.items()):
        if key not in ["source", "feature", "score", "frame"]:
            attributes.append('{} "{}"'.format(key, val))
    return "{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}".format(
        ("." if record.reference is None else record.reference.name),
        ("." if "source" not in record.annot else record.annot["source"]),
        ("." if "feature" not in record.annot else record.annot["feature"]),
        record.start,
        record.end,
        ("." if "score" not in record.annot else record.annot["score"]),
        ("." if record.strand is None else record.strand),
        ("." if "frame" not in record.annot else record.annot["frame"]),
        "; ".join(attributes)
    )

def toBracketTree(node):
    node_btree = ""
    if issubclass(node.__class__, RegionTree):
        node_btree = "{}:{}".format(node.__class__.__name__, node.getCoordinatesStr())
        if len(node.children) > 0:
            children_btree = []
            for child in node.children:
                children_btree.append(toBracketTree(child))
            node_btree += "({})".format(",".join(children_btree))
        if issubclass(node.__class__, Transcript) and len(node.proteins) > 0:
            prot_btree = []
            for prot in node.proteins:
                prot_btree.append(toBracketTree(prot))
            node_btree += "@({})".format(",".join(prot_btree))
    else:
        children_btree = []
        for child in node:
            children_btree.append(toBracketTree(child))
        node_btree = "({})".format(",".join(children_btree))
    return node_btree


########################################################################
#
# MAIN
#
########################################################################
if __name__ == "__main__":
    unittest.main()
    ###### tester UTRs
    ###### tester prot inclus dans autre



# -*- coding: utf-8 -*-
"""Classes and functions for reading/writing MSI annotations files."""

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2018 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.0.1'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

import json
from anacore.sv import HashedSVIO
from anacore.msi import Status, LocusRes, MSILocus


class MSIAnnot(HashedSVIO):
    def __init__(self, filepath, mode="r"):
        """
        Build and return an instance of HashedSVIO.

        :param filepath: The filepath.
        :type filepath: str.
        :param mode: Mode to open the file ('r', 'w', 'a').
        :type mode: str.
        :param separator: Separator used between values.
        :type separator: str.
        :param title_starter: The string used to introduce the title line.
        :type title_starter: str.
        :return: The new instance.
        :rtype: HashedSVIO
        """
        super().__init__(filepath, mode, separator="\t")
        valid_titles = ["sample", "locus_position", "method_id", "key", "value", "type"]
        if mode == "w":
            self.titles = valid_titles
        elif self.titles != valid_titles:
            raise Exception('The column present in "{}" are invalid they must be: {}.'.format(filepath, valid_titles))

    def _parseLine(self):
        """
        Return a structured record from the current line.

        :return: The record defined by the current line.
        :rtype: dict
        """
        record = super()._parseLine()
        record["value"] = getCastedValue(record["value"], record["type"])
        return record

    def recordToLine(self, record):
        """
        Return the record in SV format.

        :param record: The record to process.
        :type record: dict.
        :return: The SV line corresponding to the record.
        :rtype: str
        """
        fields = []
        for curr_title in self.titles:
            val = record[curr_title]
            if curr_title == "value" and record["type"] == "json":
                val = json.dump(val)
            else:
                val = str(val)
            fields.append(val)
        line = self.separator.join(fields)
        return(line)


def getCastedValue(value, type_name):
    casted_value = value
    if type_name != "" and type_name != "str":
        if type_name == "int":
            casted_value = int(value)
        elif type_name == "float":
            casted_value = float(value)
        elif type_name == "bool":
            casted_value = (value.lower() in ["1", "true"])
        elif type_name == "json":
            casted_value = json.loads(value)
        else:
            raise ValueError('The type "{}" cannot be used in getCastedValue.'.format(type_name))
    return casted_value


def getLocusAnnotDict(in_annot):
    data_by_spl = dict()
    with MSIAnnot(in_annot) as FH_in:
        for record in FH_in:
            # Add sample
            if record["sample"] not in data_by_spl:
                data_by_spl[record["sample"]] = dict()
            data_by_locus = data_by_spl[record["sample"]]
            # Add locus
            if record["locus_position"] not in data_by_locus:
                data_by_locus[record["locus_position"]] = dict()
            data_by_res = data_by_locus[record["locus_position"]]
            # Add result method
            if record["method_id"] not in data_by_res:
                data_by_res[record["method_id"]] = dict()
            data_by_key = data_by_res[record["method_id"]]
            # Add data
            data_by_key[record["key"]] = record["value"]
    return data_by_spl


def addLociResToSpl(msi_spl, data_by_locus, res_cls=LocusRes):
    for locus_id, data_by_res in data_by_locus.items():
        # Add locus
        if locus_id not in msi_spl.loci:
            msi_spl.addLocus(
                MSILocus(locus_id)
            )
        msi_locus = msi_spl.loci[locus_id]
        # Add result and data
        addLociResult(msi_locus, data_by_res, res_cls)


def addLociResult(msi_locus, data_by_res, res_cls):
    for result_id, data_by_key in data_by_res.items():
        # Add result
        if result_id not in msi_locus.results:
            msi_locus.results[result_id] = res_cls(Status.none)
        locus_res = msi_locus.results[result_id]
        # Add data
        for key, value in data_by_key.items():
            if key != "result_id":
                if key in locus_res.__dict__ and key != "data":  # The key correspond to an attibute
                    setattr(locus_res, key, value)
                else:  # The key correspond to a data
                    locus_res.data[key] = value



# -*- coding: utf-8 -*-
"""Classes and functions for reading/writing/processing MAF files."""

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2018 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.2.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

from anacore.sv import HashedSVIO


def getName(record):
    """
    Return an unique name to identified the variant.

    :param record: A MAF record.
    :type record: dict
    :return: The variant name.
    :rtype: str
    """
    return "{}:{}={}/{}".format(record["Chromosome"], record["Start_Position"], record["Reference_Allele"], record["Tumor_Seq_Allele2"])


class MAFIO(HashedSVIO):
    """Class to read and write a file in Mutation Annotation Format (see https://docs.gdc.cancer.gov/Data/File_Formats/MAF_Format)."""

    def __init__(self, filepath, mode="r"):
        """
        Build and return an instance of MAFIO.

        :param filepath: The filepath.
        :type filepath: str
        :param mode: Mode to open the file ('r', 'w', 'a').
        :type mode: str
        :return: The new instance.
        :rtype: MAFIO
        """
        self.version = "2.4"
        super().__init__(filepath, mode, "\t")

    def _parseHeader(self):
        """Parse MAF header to set the attribute titles."""
        if self.current_line_nb == 0:
            self.current_line = self.file_handle.readline()
            self.version = self.current_line.split(" ")[1]
            self.current_line_nb += 1
            while not self.current_line.startswith("Hugo_Symbol"):
                self.current_line = self.file_handle.readline().rstrip()
                self.current_line_nb += 1
        self.titles = [elt.strip() for elt in self.current_line.split(self.separator)]

    def isRecordLine(self, line):
        """
        Return True if the line corresponds to a record (it is not a comment or an header line).

        :param line: The evaluated line.
        :type line: str.
        :return: True if the line corresponds to a record.
        :rtype: bool
        """
        is_record = True
        if line.startswith("#") or line.startswith("Hugo_Symbol"):
            is_record = False
        return is_record

    def write(self, record):
        """
        Write record line in file.

        :param record: The record.
        :type record: dict
        """
        if self.current_line_nb == 0 and self.titles is not None:
            self.file_handle.write(
                "#version {}\n{}\n".format(
                    self.version,
                    self.separator.join(self.titles)
                )
            )
            self.current_line_nb += 2
        self.file_handle.write(self.recordToLine(record) + "\n")
        self.current_line_nb += 1



# -*- coding: utf-8 -*-
"""Classes and funcions for reading/writing text files."""

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2017 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.2.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

import os
import gzip


def isEmpty(path):
    """
    Return True if the file does not exists or is empty.

    :param path: Path to processed file.
    :type path: str
    :return: True if the file does not exists or is empty.
    :rtype: bool
    """
    is_empty = True
    if os.path.exists(path):
        if not isGzip(path):
            is_empty = os.path.getsize(path) == 0
        else:
            with gzip.open(path, 'rt') as FH:
                is_empty = FH.seek(0, 2) == 0
    return is_empty


def isGzip(path):
    """
    Return True if the file is gziped.

    :param path: Path to processed file.
    :type path: str
    :return: True if the file is gziped.
    :rtype: bool
    """
    is_gzip = None
    FH_input = gzip.open(path)
    try:
        FH_input.readline()
        is_gzip = True
    except Exception:
        is_gzip = False
    finally:
        FH_input.close()
    return is_gzip


class AbstractFile(object):
    """Class inherited by classes to read and write text file."""

    def __init__(self, filepath, mode="r"):
        """
        Build and return an instance of AbstractFile.

        :param filepath: The filepath.
        :type filepath: str
        :param mode: Mode to open the file ('r', 'w', 'a').
        :type mode: str
        :return: The new instance.
        :rtype: abstractFile.AbstractFile
        """
        self.filepath = filepath
        self.mode = mode
        if (mode in ["w", "a"] and filepath.endswith('.gz')) or (mode not in ["w", "a"] and isGzip(filepath)):
            self.file_handle = gzip.open(filepath, mode + "t")
        else:
            self.file_handle = open(filepath, mode)
        self.current_line_nb = 0
        self.current_line = None

    def close(self):
        """Closes file handle."""
        if hasattr(self, 'file_handle') and self.file_handle is not None:
            self.file_handle.close()
            self.file_handle = None
            self.current_line_nb = None
            self.current_line = None

    def closed(self):
        """
        Return True if the file is closed.

        :retrun: True if the file is closed.
        :rtype: bool
        """
        return self.file_handle.closed()

    def __del__(self):
        self.close()

    def __enter__(self):
        return(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def isRecordLine(self, line):
        """
        Return True if the line corresponds to a record (it is not a comment or an header line).

        :param line: The evaluated line.
        :type line: str.
        :return: True if the line corresponds to a record.
        :rtype: bool
        """
        return True

    def __iter__(self):
        for line in self.file_handle:
            self.current_line = line.rstrip("\n")
            self.current_line_nb += 1
            if not self.isRecordLine(self.current_line):  # Skip non-record line
                continue
            record = None
            try:
                record = self._parseLine()
            except Exception:
                raise IOError('The line {} in "{}" cannot be parsed by {}. Line content: {}'.format(self.current_line_nb, self.filepath, self.__class__.__name__, self.current_line))
            else:
                yield record

    def read(self):
        """
        Return all the record from the file.

        :return: The list of records.
        :rtype: list
        """
        return [record for record in self]



# -*- coding: utf-8 -*-
"""Classes and functions for reading/writing mSINGS's outputs."""

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2018 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.3.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

import gzip
from anacore.abstractFile import isGzip
from anacore.sv import HashedSVIO
from anacore.msi import MSILocus, LocusRes, MSISample, MSISplRes, Status


class MSINGSAnalysis(HashedSVIO):
    """Manage output file produced by the command "msi analyzer" of mSINGS (https://bitbucket.org/uwlabmed/msings)."""

    def __init__(self, filepath, mode="r"):
        """
        Return the new instance of MSINGSAnalysis.

        :param filepath: The filepath.
        :type filepath: str
        :param mode: Mode to open the file ('r', 'w', 'a').
        :type mode: str
        """
        super().__init__(filepath, mode, "\t")
        if mode == "w":
            self.titles = ["Position", "Name", "Average_Depth", "Number_of_Peaks", "Standard_Deviation", "IndelLength:AlleleFraction:SupportingCalls"]

    def _parseLine(self):
        """
        Return a structured record from the current line.

        :return: The record described by the current line.
        :rtype: MSILocus
        """
        record = super()._parseLine()
        peaks = record["IndelLength:AlleleFraction:SupportingCalls"].split(" ")
        nb_by_length = {}
        if len(peaks) == 1 and peaks[0] == "0:0.0:0":
            peaks = []
        else:
            for idx, curr_peak in enumerate(peaks):
                indel_length, AF, DP = curr_peak.split(":")
                peaks[idx] = {
                    "indel_length": int(indel_length),
                    "AF": float(AF),
                    "DP": int(DP)
                }
                nb_by_length[int(indel_length)] = int(DP)

        return MSILocus.fromDict({
            "position": record["Position"],
            "name": record["Name"],
            "results": {
                "MSINGS": {
                    "_class": "LocusResDistrib",
                    "status": Status.undetermined,
                    "data": {
                        "avg_depth": record["Average_Depth"],
                        "nb_peaks": record["Number_of_Peaks"],
                        "peaks": peaks,
                        "std_dev": record["Standard_Deviation"],
                        "nb_by_length": nb_by_length
                    }
                }
            }
        })

    def recordToLine(self, record):
        """
        Return the record in SV format.

        :param record: The record containing a result coming from mSINGS.
        :type record: MSILocus
        :return: The SV line corresponding to the record.
        :rtype: str
        """
        formatted_record = {
            "Position": record.position,
            "Name": record.name,
            "Average_Depth": record.results["MSINGS"].data["avg_depth"],
            "Number_of_Peaks": record.results["MSINGS"].data["nb_peaks"],
            "Standard_Deviation": record.results["MSINGS"].data["std_dev"]
        }
        formatted_record["IndelLength:AlleleFraction:SupportingCalls"] = " ".join(
            ["{}:{}:{}".format(curr_peak["indel_length"], curr_peak["AF"], curr_peak["DP"]) for curr_peak in record.results["MSINGS"].data["peaks"]]
        )
        return super().recordToLine(formatted_record)


class MSINGSReport(object):
    """Manage output file produced by the command "msi count_msi_samples" of mSINGS (https://bitbucket.org/uwlabmed/msings)."""

    def __init__(self, filepath):
        """
        Return the new instance of MSINGSReport.

        :param filepath: The filepath.
        :type filepath: str
        """
        self.filepath = filepath
        self.samples = dict()
        self.loci = list()
        self.method_name = "MSINGS"
        self.parse()

    def _parseFileHandle(self, FH):
        """
        Parse file referenced by the file handle FH and store information in self.

        :param FH: The file handle for the filepath.
        :type FH: TextIOWrapper
        """
        # Parse general information
        samples = [elt.strip() for elt in FH.readline().split("\t")[1:]]
        nb_unstable = [int(elt.strip()) for elt in FH.readline().split("\t")[1:]]
        nb_evaluated = [int(elt.strip()) for elt in FH.readline().split("\t")[1:]]
        line = FH.readline()
        if not line.startswith("msing_score"):
            scores = [None for curr_spl in samples]
        else:
            scores = [elt.strip() for elt in line.split("\t")][1:]
            for idx, elt in enumerate(scores):
                curr_score = None
                if elt != "":
                    curr_score = float(elt)
                scores[idx] = curr_score
            line = FH.readline()  # To next line
        status = [elt.strip() for elt in line.split("\t")][1:]
        for idx, elt in enumerate(status):
            curr_status = Status.undetermined
            if nb_evaluated[idx] > 0:
                if elt == "NEG":
                    curr_status = Status.stable
                elif elt == "POS":
                    curr_status = Status.unstable
            status[idx] = curr_status
        for spl_idx, curr_spl in enumerate(samples):
            if status[spl_idx] == Status.stable:
                """
                Convert mSINGS score (ratio unstable / determined) to a confidence score.
                Nb loci unstables:    0    1    2    3    4
                mSINGS score:         0 0.25 0.50 0.75    1
                confi unstable:       0 0.25 0.50 0.75    1
                confi stable:         1 0.75 0.50 0.25    0
                """
                scores[spl_idx] = 1 - scores[spl_idx]
            spl_res = MSISplRes(status[spl_idx], scores[spl_idx], self.method_name)
            self.samples[curr_spl] = MSISample(curr_spl, None, {self.method_name: spl_res})
        # Parse loci information
        for curr_line in FH:
            fields = [elt.strip() for elt in curr_line.split("\t")]
            curr_locus = fields[0]
            self.loci.append(curr_locus)
            for idx, curr_val in enumerate(fields[1:]):
                curr_spl = samples[idx]
                loci_res = None
                if curr_val == "":
                    loci_res = LocusRes(Status.undetermined)
                elif curr_val == "1":
                    loci_res = LocusRes(Status.unstable)
                else:
                    loci_res = LocusRes(Status.stable)
                self.samples[curr_spl].addLocus(
                    MSILocus(curr_locus, None, {self.method_name: loci_res})
                )

    def parse(self):
        """Parse file and store information in self."""
        if isGzip(self.filepath):
            with gzip.open(self.filepath, "rt") as FH:
                self._parseFileHandle(FH)
        else:
            with open(self.filepath, "r") as FH:
                self._parseFileHandle(FH)



# -*- coding: utf-8 -*-
"""Classes and functions for manipulating/processing genomic regions defines on references sequences (Gene, Transcripts, Proteins, Regulation sequences, ...)."""

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2018 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.7.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

from anacore.region import RegionTree, RegionList
from anacore.sequenceIO import getStrandedSeqFromPos


class Gene(RegionTree):
    pass


class Exon(RegionTree):
    pass


class Intron(RegionTree):
    def __init__(self, start=None, end=None, strand=None, reference=None, name=None, annot=None, parent=None, children=None):
        """
        Build and return an instance of Intron.

        :param start: The start position on the reference. This position is 1-based and ascending (start <= end).
        :type start: int
        :param end: The end position on the reference. This position is 1-based and ascending (start <= end). [Default: start]
        :type end: int
        :param strand: The strand of the instance ("+" or "-").
        :type strand: str
        :param reference: The region object or the region name of the reference.
        :type reference: region.Region | str
        :param name: The name of the region.
        :type name: str
        :param annot: The annotations of the region.
        :type annot: dict
        :param parent: The parent region (example: gene for a transcript).
        :type parent: region.RegionTree
        :param children: The list of sub-regions (example: exons in transcript).
        :type children: region.RegionList
        :return: The new instance.
        :rtype: region.Intron
        """
        RegionTree.__init__(self, start, end, strand, reference, name, annot, None, children)
        self.parent = parent  # Prevent add of this intron as child of transcript


class CDS(RegionTree):
    pass


class Transcript(RegionTree):
    """Class to manage a transcript region."""

    def __init__(self, start=None, end=None, strand=None, reference=None, name=None, annot=None, parent=None, children=None, proteins=None):
        """
        Build and return an instance of Transcript.

        :param start: The start position on the reference. This position is 1-based and ascending (start <= end).
        :type start: int
        :param end: The end position on the reference. This position is 1-based and ascending (start <= end). [Default: start]
        :type end: int
        :param strand: The strand of the instance ("+" or "-").
        :type strand: str
        :param reference: The region object or the region name of the reference.
        :type reference: region.Region | str
        :param name: The name of the region.
        :type name: str
        :param annot: The annotations of the region.
        :type annot: dict
        :param parent: The gene.
        :type parent: region.RegionTree
        :param children: The list of exons.
        :type children: region.RegionList
        :param proteins: The list of proteins produced on transcript. It can exists several proteins on transcript: operon and readthrough.
        :type proteins: region.RegionList
        :return: The new instance.
        :rtype: genomicRegion.Transcript
        """
        RegionTree.__init__(self, start, end, strand, reference, name, annot, parent, children)
        self._proteins = RegionList()
        self.proteins = proteins

    @property
    def proteins(self):
        return self._proteins

    @proteins.setter
    def proteins(self, proteins):
        """
        Change proteins linked with transcript. Before the add of the new proteins the old are unlinked (prot.transcript is set to None).

        :param proteins: The new linked proteins.
        :type proteins: list
        """
        # Remove previous proteins
        for idx_prot, prot in enumerate(self.proteins):
            prot.transcript = None  # Remove link in protein
        self._proteins = RegionList()  # Remove link in transcript
        # Add new proteins
        if proteins is not None:
            for curr_prot in proteins:
                curr_prot.transcript = self

    def addProtein(self, protein):
        """
        Add link with a protein.

        :param protein: The protein to add.
        :type protein: genomicRegion.Protein
        """
        protein.transcript = self

    def delProtein(self, protein):
        """
        Remove link with a protein.

        :param protein: The protein to remove.
        :type protein: genomicRegion.Protein
        """
        protein.transcript = None

    def length(self):
        """
        Return length of the transcript (sum of exons lengths).

        :return: The length of the transcript.
        :rtype: int
        """
        length = 0
        for curr_exon in self.children:
            length += curr_exon.length()
        return length

    def getPosOnRef(self, transcript_pos):
        """
        Return coordinate on reference sequence from the coordinate on transcript.

        :param transcript_pos: Coordinate on transcript (1-based).
        :type transcript_pos: int.
        :return: The coordinate on reference sequence (1-based).
        :rtype: int
        """
        exon, region_pos = self.getSubFromRegionPos(transcript_pos)
        return exon.getPosOnRef(region_pos)

    def getPosOnRegion(self, chr_pos):
        """
        Return coordinate on transcript from the coordinate on reference sequence.

        :param chr_pos: The coordinate on reference sequence (1-based).
        :type chr_pos: int
        :return: Coordinate on region (1-based).
        :rtype: int
        """
        if chr_pos < self.start or chr_pos > self.end:
            raise ValueError("The position {} is out of transcript {}.".format(chr_pos, self))
        if self.strand is None:
            raise Exception("Cannot return region position from reference position because the strand is None ({}).".format(
                self
            ))
        transcript_pos = 0
        idx = 0
        found = False
        if self.strand == "+":
            while not found:
                curr_exon = self.children[idx]
                if chr_pos > curr_exon.end:  # Next exon
                    transcript_pos += curr_exon.length()
                    idx += 1
                elif chr_pos >= curr_exon.start:  # In exon
                    transcript_pos += chr_pos - curr_exon.start + 1
                    found = True
                else:  # In previous intron
                    raise Exception("Position {}:{} is in intron of {}.".format(self.reference.name, chr_pos, self))
        else:  # Strand -
            while not found:
                curr_exon = self.children[idx]
                if chr_pos < curr_exon.start:  # Next exon
                    transcript_pos += curr_exon.length()
                    idx += 1
                elif chr_pos <= curr_exon.end:  # In exon
                    transcript_pos += curr_exon.end - chr_pos + 1
                    found = True
                else:  # In previous intron
                    raise Exception("Position {}:{} is in intron of {}.".format(self.reference.name, chr_pos, self))
        return transcript_pos

    def getSubFromRegionPos(self, transcript_pos):
        """
        Return exon and coordinate on its from coordinate on transcript. Coordinate on exon is stranded:

            * pos 2 in region test:10-20(+) correspond to ref position 12
            * pos 2 in region test:10-20(-) correspond to ref position 18

        :param transcript_pos: The coordinate on transcript (1-based).
        :type transcript_pos: int
        :return: Exon and coordinate on exon (1-based).
        :rtype: (genomicRegion.Exon, int)
        """
        if transcript_pos > self.length():
            raise ValueError("The position {} is out of transcript {}.".format(transcript_pos, self))
        walk_pos = 0
        region_idx = 0
        child = None
        while child is None:
            curr_child_length = self.children[region_idx].length()
            if walk_pos <= transcript_pos and (walk_pos + curr_child_length) >= transcript_pos:
                child = self.children[region_idx]
            else:
                walk_pos += curr_child_length
                region_idx += 1
        pos_on_child = transcript_pos - walk_pos
        return child, pos_on_child

    def getSubFromRefPos(self, chr_pos):
        """
        Return exon on intron where the chr_pos is located.

        :param chr_pos: The coordinate on reference sequence (1-based).
        :type chr_pos: int
        :return: The index of the sub-region containing the chr_pos (1-based) and the object of this sub-region.
        :rtype: int, (Intron | Exon)
        """
        if self.strand is None:
            raise Exception("Cannot return a region position from the reference position because the strand is None ({}).".format(
                self
            ))
        if chr_pos < self.start or chr_pos > self.end:
            raise ValueError("The position {} is out of transcript {}.".format(chr_pos, self))
        sub_region_idx = None
        sub_region = None
        exons = self.children
        exon_idx = 0
        if self.strand == "+":
            while chr_pos > exons[exon_idx].end:
                exon_idx += 1
            if chr_pos < exons[exon_idx].start:
                intron_start = exons[exon_idx - 1].end + 1
                intron_end = exons[exon_idx].start - 1
                sub_region_idx = exon_idx
                sub_region = Intron(intron_start, intron_end, self.strand, self.reference, "intron_{}".format(sub_region_idx), {"siblings_idx": sub_region_idx}, parent=self)
            else:
                sub_region_idx = exon_idx + 1
                sub_region = exons[exon_idx]
        else:
            while chr_pos < exons[exon_idx].start:
                exon_idx += 1
            if chr_pos > exons[exon_idx].end:
                intron_end = exons[exon_idx - 1].start - 1
                intron_start = exons[exon_idx].end + 1
                sub_region_idx = exon_idx
                sub_region = Intron(intron_start, intron_end, self.strand, self.reference, "intron_{}".format(sub_region_idx), {"siblings_idx": sub_region_idx}, parent=self)
            else:
                sub_region_idx = exon_idx + 1
                sub_region = exons[exon_idx]
        return sub_region, sub_region_idx


class Protein(RegionTree):
    """Class to manage a protein region."""

    def __init__(self, start=None, end=None, strand=None, reference=None, name=None, annot=None, parent=None, children=None, transcript=None):
        """
        Build and return an instance of Protein.

        :param start: The start position on the reference. This position is 1-based and ascending (start <= end).
        :type start: int
        :param end: The end position on the reference. This position is 1-based and ascending (start <= end). [Default: start]
        :type end: int
        :param strand: The strand of the instance ("+" or "-").
        :type strand: str
        :param reference: The region object or the region name of the reference.
        :type reference: region.Region | str
        :param name: The name of the region.
        :type name: str
        :param annot: The annotations of the region.
        :type annot: dict
        :param parent: The gene.
        :type parent: region.RegionTree
        :param children: The list of CDS.
        :type children: region.RegionList
        :param transcript: The transcript corresponding to the protein.
        :type transcript: region.Transcript
        :return: The new instance.
        :rtype: genomicRegion.Protein
        """
        RegionTree.__init__(self, start, end, strand, reference, name, annot, parent, children)
        self._transcript = None
        self.transcript = transcript

    @property
    def transcript(self):
        return self._transcript

    @transcript.setter
    def transcript(self, transcript):
        """
        Add link between self and his transcript.

        :param transcript: The transcript to link.
        :type transcript: genomicRegion.Transcript
        """
        self._transcript = transcript
        # Manage link in transcript
        if transcript is not None:
            # Add link in transcript
            if self not in transcript.proteins:
                transcript.proteins.append(self)
        else:  # New transcript is None
            # Remove link in previous transcript
            prot_idx = None
            try:
                prot_idx = transcript.proteins.index(self)
            except Exception:  # Not found
                pass
            if prot_idx is not None:
                del transcript.proteins[prot_idx]

    def length(self):
        """
        Return length of the protein in nucleotids (sum of CDS lengths).

        :return: The length of the protein.
        :rtype: int
        """
        length = 0
        for curr_cds in self._getCDS():
            length += curr_cds.length()
        return length

    def aaLength(self):
        """
        Return length of the protein in amino acids (sum of CDS lengths).

        :return: The length of the protein.
        :rtype: int
        """
        return int(self.length() / 3)

    def getPosOnRef(self, protein_pos, codon_pos):
        """
        Return coordinate on reference sequence from the coordinate on protein.

        :param protein_pos: Coordinate of the amino acid on protein (1-based).
        :type protein_pos: int.
        :param codon_pos: Coordinate on the codon (1-based).
        :type codon_pos: int.
        :return: The coordinate on reference sequence (1-based).
        :rtype: int
        """
        cds, cds_pos = self.getSubFromRegionPos(protein_pos, codon_pos)
        return cds.getPosOnRef(cds_pos)

    def getPosOnRegion(self, chr_pos):
        """
        Return coordinate on protein from the coordinate on reference sequence.

        :param chr_pos: The coordinate on reference sequence (1-based).
        :type chr_pos: int
        :return: Coordinate of the amino acid on protein (1-based) and the coordinate on the codon (1-based). Return None if coordinates are in UTRs or in introns.
        :rtype: (int, int) | None
        """
        if chr_pos < self.start or chr_pos > self.end:
            raise ValueError("The position {} is out of protein {}.".format(chr_pos, self))
        chained_cds_pos = self.getNtPosFromRefPos(chr_pos)
        if chained_cds_pos is None:
            return [None, None]
        return(
            int((chained_cds_pos - 1) / 3) + 1,  # aa position (1-based)
            ((chained_cds_pos - 1) % 3) + 1  # position in aa (1-based)
        )

    def getNtPosFromRegionPos(self, protein_pos, codon_pos):
        """
        Return nucleotids coordinate on protein from the amino acids coordinate.

        :param protein_pos: Coordinate of the amino acid on protein (1-based).
        :type protein_pos: int.
        :param codon_pos: Coordinate on the codon (1-based).
        :type codon_pos: int.
        :return: The nucleotids coordinate on protein (1-based).
        :rtype: int
        """
        if protein_pos > self.aaLength():
            raise ValueError("The position {} is out of protein {}.".format(protein_pos, self))
        return ((protein_pos - 1) * 3) + codon_pos

    def getNtPosFromRefPos(self, chr_pos):
        """
        Return nucleotids coordinate on protein from coordinate on reference sequence.

        :param chr_pos: The coordinate on reference sequence (1-based).
        :type chr_pos: int
        :return: The nucleotids coordinate on protein (1-based). Return None if coordinates are in UTRs or in introns.
        :rtype: int | None
        """
        if self.strand is None:
            raise Exception("Cannot return a region position from the reference position because the strand is None ({}).".format(
                self
            ))
        chained_cds_pos = None
        cds = self._getCDS()
        cds_idx = 0
        if cds[cds_idx].strand == "+":
            if chr_pos <= cds[-1].end and chr_pos >= cds[0].start:
                chained_cds_pos = 0
                while chr_pos > cds[cds_idx].end:
                    chained_cds_pos += cds[cds_idx].length()
                    cds_idx += 1
                chained_cds_pos += chr_pos - cds[cds_idx].start + 1
        else:
            if chr_pos <= cds[0].end and chr_pos >= cds[-1].start:
                chained_cds_pos = 0
                while chr_pos < cds[cds_idx].start:
                    chained_cds_pos += cds[cds_idx].length()
                    cds_idx += 1
                chained_cds_pos += cds[cds_idx].end - chr_pos + 1
        return chained_cds_pos

    def getSubFromRegionPos(self, protein_pos, codon_pos):
        """
        Return CDS and coordinate on its from coordinate on protein. The coordinate on CDS is stranded:

            * With region test:10-20(+) the pos 2 correspond to ref position 12
            * With region test:10-20(-) the pos 2 correspond to ref position 18

        :param protein_pos: Coordinate of the amino acid on protein (1-based).
        :type protein_pos: int.
        :param codon_pos: Coordinate on the codon (1-based).
        :type codon_pos: int.
        :return: CDS and coordinate on CDS (1-based).
        :rtype: genomicRegion.CDS, int
        """
        chained_cds_pos = self.getNtPosFromRegionPos(protein_pos, codon_pos)
        cds = self._getCDS()
        walk_pos = 0
        region_idx = 0
        child = None
        while child is None:
            curr_child_length = cds[region_idx].length()
            if walk_pos <= chained_cds_pos and (walk_pos + curr_child_length) >= chained_cds_pos:
                child = cds[region_idx]
            else:
                walk_pos += curr_child_length
                region_idx += 1
        pos_on_child = chained_cds_pos - walk_pos
        return child, pos_on_child

    def getCDSFromTranscript(self):
        """
        Return CDS of the protein from the transcript and his exons. This function is used when CDS are not defined in the protein but exons and protein start and end are defined.

        :return: The list of CDS of the protein in protein strand order.
        :rtype: region.Regionlist
        """
        # Check information completion
        if self.transcript is None:
            raise Exception("A link with a transcript is required to return CDS for {}.".format(self))
        if self.start is None or self.end is None:
            raise Exception("Start and end for {} are required to return CDS from transcript {}.".format(self, self.transcript))
        # Exons to CDS
        exons = sorted(self.transcript.children, key=lambda x: (x.start, x.end))
        nb_exons = len(exons)
        idx_exon = 0
        curr_exon = exons[idx_exon]
        while self.start > curr_exon.end:
            idx_exon += 1
            curr_exon = exons[idx_exon]
        cds = RegionList()
        while curr_exon is not None and self.end >= curr_exon.start:
            cds_start = max(self.start, curr_exon.start)
            cds_end = min(self.end, curr_exon.end)
            cds.append(CDS(cds_start, cds_end, self.strand, self.reference))
            idx_exon += 1
            curr_exon = None
            if idx_exon < nb_exons:
                curr_exon = exons[idx_exon]
        # Sort by strand order
        if self.strand == "-":
            cds = RegionList(sorted(cds, key=lambda x: (x.end, x.start), reverse=True))
        # Return
        return cds

    def contains(self, eval_region):
        """
        Return True if at least one CDS has an overlap with eval_region.

        :param eval_region: The evaluated region.
        :type eval_region: Region
        :return: True if the region contains the evaluated region.
        :rtype: bool
        """
        cds = self._getCDS()
        return len(cds.getContainers(eval_region)) > 0

    def _getCDS(self):
        """
        Init self.children if necessary and return the list of CDS.

        :param eval_region: The evaluated region.
        :type eval_region: Region
        :return: The list of CDS of the protein in protein strand order.
        :rtype: anacore.region.Regionlist
        """
        if len(self.children) == 0:
            self.children = self.getCDSFromTranscript()
        return self.children

    def hasOverlap(self, eval_region):
        """
        Return True if at least one CDS has an overlap with eval_region.

        :param eval_region: The evaluated region.
        :type eval_region: Region
        :return: True if the protein has an overlap with evaluated region.
        :rtype: bool
        """
        has_overlap = False
        if self.reference.name == eval_region.reference.name:
            cds = self._getCDS()
            for curr_cds in cds:
                if not curr_cds.start > eval_region.end and not curr_cds.end < eval_region.start:
                    has_overlap = True
        return has_overlap

    def getCodonRefPos(self, aa_pos):
        """
        Return stranded positions on reference for the nt corresponding to the amino acid at provided position on protein.

        :param aa_pos: Position of the amino acid in the protein.
        :type aa_pos: int
        :return: Stranded positions on reference for the nt corresponding to the amino acid at provided position on protein.
        :rtype: list
        """
        codon_pos_on_ref = []
        for curr_pos in range(1, 4):
            codon_pos_on_ref.append(self.getPosOnRef(aa_pos, curr_pos))
        return codon_pos_on_ref

    def getCodonSeqFromProtPos(self, aa_pos, sequence_reader):
        """
        Return the codon nucleotids corresponding to the amino acid at the provided position on protein.

        :param aa_pos: Position of the amino acid in the protein.
        :type aa_pos: int
        :param sequence_reader: File handle to the reference sequences file.
        :type sequence_reader: anacore.sequenceIO.IdxFastaIO
        :return: Codon nucleotids corresponding to the amino acid at the provided position on protein.
        :rtype: str
        """
        codon_pos_on_ref = self.getCodonRefPos(aa_pos)
        return getStrandedSeqFromPos(
            self.reference.name,
            codon_pos_on_ref,
            self.strand,
            sequence_reader
        )

    def getCodonInfo(self, ref_pos, sequence_reader):
        """
        Return the amino acid position, position in codon and codon sequence for the position on reference.

        :param ref_pos: Position on the reference.
        :type ref_pos: int
        :param sequence_reader: File handle to the reference sequences file.
        :type sequence_reader: anacore.sequenceIO.IdxFastaIO
        :return: Amino acid position, position in codon and codon sequence for the position on reference.
        :rtype: (int, int, str)
        """
        # Get codon pos
        aa_pos, pos_in_codon = self.getPosOnRegion(ref_pos)
        codon_pos_on_ref = []
        for curr_pos in range(1, 4):
            if curr_pos == pos_in_codon:
                codon_pos_on_ref.append(ref_pos)
            else:
                codon_pos_on_ref.append(self.getPosOnRef(aa_pos, curr_pos))
        # Get codon seq
        codon_seq = getStrandedSeqFromPos(
            self.reference.name,
            codon_pos_on_ref,
            self.strand,
            sequence_reader
        )
        return aa_pos, pos_in_codon, codon_seq



# -*- coding: utf-8 -*-
"""Classes and functions for manipulating/processing sequences."""

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2021 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.0.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

from textwrap import wrap


class Alphabet:
    pass


class AA3LettersAlphabet(Alphabet):
    """Static methods and attributes to manage amino acid alphabet in three letters."""
    codons_by_aa = {
        "Ter": ["TAA", "TGA", "TAG"],
        "Ala": ["GCT", "GCC", "GCA", "GCG"],
        "Arg": ["CGT", "CGC", "CGA", "CGG", "AGA", "AGG"],
        "Asn": ["AAT", "AAC"],
        "Asp": ["GAT", "GAC"],
        "Cys": ["TGT", "TGC"],
        "Gln": ["CAA", "CAG"],
        "Glu": ["GAA", "GAG"],
        "Gly": ["GGT", "GGC", "GGA", "GGG"],
        "His": ["CAT", "CAC"],
        "Ile": ["ATT", "ATC", "ATA"],
        "Leu": ["CTT", "CTC", "CTA", "CTG", "TTA", "TTG"],
        "Lys": ["AAA", "AAG"],
        "Met": ["ATG"],
        "Phe": ["TTT", "TTC"],
        "Pro": ["CCT", "CCC", "CCA", "CCG"],
        "Ser": ["TCT", "TCC", "TCA", "TCG", "AGT", "AGC"],
        "Thr": ["ACT", "ACC", "ACA", "ACG"],
        "Trp": ["TGG"],
        "Tyr": ["TAT", "TAC"],
        "Val": ["GTT", "GTC", "GTA", "GTG"]
    }

    one_by_three = {
        "Ter": "*",
        "Ala": "A",
        "Arg": "R",
        "Asn": "N",
        "Asp": "D",
        "Cys": "C",
        "Gln": "Q",
        "Glu": "E",
        "Gly": "G",
        "His": "H",
        "Ile": "I",
        "Leu": "L",
        "Lys": "K",
        "Met": "M",
        "Phe": "F",
        "Pro": "P",
        "Ser": "S",
        "Thr": "T",
        "Trp": "W",
        "Tyr": "Y",
        "Val": "V",
        "Asx": "B",
        "Glx": "Z",
        "Xle": "J",
        "Sec": "U",
        "Pyl": "O",
        "Xaa": "X"
    }

    words = set(codons_by_aa.keys())

    @staticmethod
    def isValid(seq):
        """
        Return True if the sequence contains only elements of the amino acid alphabet.

        :param seq: Sequence to validate.
        :type seq: str
        :return: True if the sequence contains only elements of the amino acid alphabet.
        :rtype: bool
        """
        is_valid = True
        try:
            for aa in wrap(seq, 3):
                if aa.capitalize() not in __class__.words:
                    is_valid = False
                    break
        except Exception:
            pass
        return is_valid

    @staticmethod
    def toOneLetter(seq):
        """
        Return sequence of amino acids in on letter alphabet.

        :param seq: Sequence to convert.
        :type seq: str
        :return: Sequence of amino acids in on letter alphabet.
        :rtype: str
        """
        new_seq = ""
        for aa in wrap(seq, 3):
            new_seq += __class__.one_by_three[aa]
        return new_seq


class CodonAlphabet(Alphabet):
    """Static methods and attributes to manage codon alphabet."""
    aa_by_codon = {curr_codon: aa for aa, codons in AA3LettersAlphabet.codons_by_aa.items() for curr_codon in codons}

    words = set(aa_by_codon.keys())

    @staticmethod
    def isValid(seq):
        """
        Return True if the sequence contains only elements of the codon alphabet.

        :param seq: Sequence to validate.
        :type seq: str
        :return: True if the sequence contains only elements of the codon alphabet.
        :rtype: bool
        """
        is_valid = True
        try:
            for codon in wrap(seq, 3):
                if codon.upper() not in __class__.words:
                    is_valid = False
                    break
        except Exception:
            pass
        return is_valid

    @staticmethod
    def translate(seq):
        """
        Return amino acids sequence from codon sequence.

        :param seq: Sequence to translate.
        :type seq: str
        :return: Translated sequence.
        :rtype: str
        """
        return "".join([CodonAlphabet.aa_by_codon[codon.upper()] for codon in wrap(seq, 3)])


class DNAAlphabet(Alphabet):
    """Static methods and attributes to manage DNA alphabet."""
    complement = {
        'A': 'T', 'T': 'A', 'G': 'C', 'C': 'G', 'U': 'A', 'N': 'N',
        'a': 't', 't': 'a', 'g': 'c', 'c': 'g', 'u': 'a', 'n': 'n',
        'W': 'W', 'S': 'S', 'M': 'K', 'K': 'M', 'R': 'Y', 'Y': 'R', 'B': 'V', 'V': 'B', 'D': 'H', 'H': 'D',
        'w': 'w', 's': 's', 'm': 'k', 'k': 'm', 'r': 'y', 'y': 'r', 'b': 'v', 'v': 'b', 'd': 'h', 'h': 'd'
    }

    words = set(complement.keys())

    @staticmethod
    def isValid(seq):
        """
        Return True if the sequence contains only elements of the DNA alphabet.

        :param seq: Sequence to validate.
        :type seq: str
        :return: True if the sequence contains only elements of the DNA alphabet.
        :rtype: bool
        """
        is_valid = True
        for nt in seq:
            if nt not in __class__.words:
                is_valid = False
                break
        return is_valid

    @staticmethod
    def revCom(seq):
        """
        Return the reverse complement.

        :param seq: Sequence to reverse complement.
        :type seq: str
        :return: The reverse complement.
        :rtype: str
        """
        return "".join([__class__.complement[base] for base in seq[::-1]])


class RNAAlphabet(Alphabet):
    """Static methods and attributes to manage RNA alphabet."""
    complement = {
        'A': 'U', 'T': 'A', 'G': 'C', 'C': 'G', 'U': 'A', 'N': 'N',
        'a': 'u', 't': 'a', 'g': 'c', 'c': 'g', 'u': 'a', 'n': 'n',
        'W': 'W', 'S': 'S', 'M': 'K', 'K': 'M', 'R': 'Y', 'Y': 'R', 'B': 'V', 'V': 'B', 'D': 'H', 'H': 'D',
        'w': 'w', 's': 's', 'm': 'k', 'k': 'm', 'r': 'y', 'y': 'r', 'b': 'v', 'v': 'b', 'd': 'h', 'h': 'd'
    }

    words = set(complement.keys())

    @staticmethod
    def isValid(seq):
        """
        Return True if the sequence contains only elements of the RNA alphabet.

        :param seq: Sequence to validate.
        :type seq: str
        :return: True if the sequence contains only elements of the RNA alphabet.
        :rtype: bool
        """
        is_valid = True
        for nt in seq:
            if nt not in __class__.words:
                is_valid = False
                break
        return is_valid

    @staticmethod
    def revCom(seq):
        """
        Return the reverse complement.

        :param seq: Sequence to reverse complement.
        :type seq: str
        :return: The reverse complement.
        :rtype: str
        """
        return "".join([__class__.complement[base] for base in seq[::-1]])


class Sequence:
    def __init__(self, id, string, description=None, quality=None):
        """
        Build and return an instance of Sequence.

        :param id: Id of the sequence.
        :type id: str
        :param string: Sequence of the sequence.
        :type string: str
        :param description: The sequence description.
        :type description: str
        :param quality: The quality of the sequence (same length as string).
        :type quality: str
        :return: The new instance.
        :rtype: Sequence
        """
        self.id = id
        self.description = description
        self.string = string
        self.quality = quality

    def dnaRevCom(self):
        """
        Return the sequence corresponding to the DNA reverse complement.

        :return: The reverse complement.
        :rtype: Sequence
        """
        return Sequence(
            self.id,
            DNAAlphabet.revCom(self.string),
            self.description,
            (None if self.quality is None else self.quality[::-1])
        )

    def rnaRevCom(self):
        """
        Return the sequence corresponding to the RNA reverse complement.

        :return: The reverse complement.
        :rtype: Sequence
        """
        return Sequence(
            self.id,
            RNAAlphabet.revCom(self.string),
            self.description,
            (None if self.quality is None else self.quality[::-1])
        )



#!/usr/bin/env python
import os
import re
from setuptools import find_packages, setup


def get_long_description():
    content = ""
    with open("README.md", encoding='utf-8') as FH:
        content = FH.read()
    return content


def get_version():
    version = None
    notes_filepath = "RELEASES_NOTES.md"
    if os.path.exists(notes_filepath):
        with open(notes_filepath) as FH:
            first_line = FH.readline()
            version = re.search(r"^\#\s+.+\s+(.+)\s+\[", first_line).groups()[0]  # Example: "# v2.5.0 [DEV]"
    return version


def load_requirements(path):
    requirements = []
    with open(path) as FH:
        requirements = [elt.strip().replace(" ", "") for elt in FH]
    return requirements


setup(
    name='anacore',
    version=get_version(),
    description='Libraries for managing standard file formats and objects from NGS.',
    author='Frederic Escudie',
    author_email='escudie.frederic@iuct-oncopole.fr',
    license='GNU GPL v3',
    packages=find_packages(),
    install_requires=load_requirements("requirements.txt"),
    url='https://github.com/bialimed/anacore',
    python_requires='>=3.5',
    keywords='bio NGS',
    #long_description_content_type='text/markdown',
    #long_description=get_long_description()
)



#!/usr/bin/env python3

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2019 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.6.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

import os
import sys
import gzip
import uuid
import filecmp
import tempfile
import unittest

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
PACKAGE_DIR = os.path.dirname(TEST_DIR)
sys.path.append(PACKAGE_DIR)

from anacore.sequence import RNAAlphabet, Sequence
from anacore.sequenceIO import IdxFastaIO, FastaIO, FastqIO, getStrandedSeqFromPos


########################################################################
#
# FUNCTIONS
#
########################################################################
class TestFastaIO(unittest.TestCase):
    def setUp(self):
        tmp_folder = tempfile.gettempdir()
        unique_id = str(uuid.uuid1())

        # Temporary files
        self.tmp_multi_line = os.path.join(tmp_folder, unique_id + "_multi.fasta")
        self.tmp_multi_line_gz = os.path.join(tmp_folder, unique_id + "_multi.fasta.gz")
        self.tmp_mono_line = os.path.join(tmp_folder, unique_id + "_mono.fasta")
        self.tmp_out = os.path.join(tmp_folder, unique_id + "_out.fasta")

        # Expected records
        self.expected_rec = [
            Sequence("seq1", "ATAGATAGCATCCCCCCNATACATAGATAGCATCCCCCCNATACATAGATAGCATCCCCCCNATAC", "test description"),
            Sequence("seq2", ""),
            Sequence("seq3", "ATGAAAAAAAAAAAAANTGATGAAAAAAAAAAAAANTG", "test description 2"),
            Sequence("seq4", "", "trimmed")
        ]

        # Create multi line
        content = """>seq1 test description
ATAGATAGCATCCCCCCNATAC
ATAGATAGCATCCCCCCNATAC
ATAGATAGCATCCCCCCNATAC
>seq2

>seq3 test description 2
ATGAAAAAAAAAAAAANTG
ATGAAAAAAAAAAAAANTG
>seq4 trimmed

"""
        with open(self.tmp_multi_line, "w") as FH_out:
            FH_out.write(content)

        # Create mono line
        content = """>seq1 test description
ATAGATAGCATCCCCCCNATACATAGATAGCATCCCCCCNATACATAGATAGCATCCCCCCNATAC
>seq2

>seq3 test description 2
ATGAAAAAAAAAAAAANTGATGAAAAAAAAAAAAANTG
>seq4 trimmed

"""
        with open(self.tmp_mono_line, "w") as FH_out:
            FH_out.write(content)
        with gzip.open(self.tmp_multi_line_gz, "wt") as FH_out:
            FH_out.write(content)

    def testNbSeq(self):
        nb_seq = FastaIO.nbSeq(self.tmp_mono_line)
        self.assertEqual(4, nb_seq)
        nb_seq = FastaIO.nbSeq(self.tmp_multi_line)
        self.assertEqual(4, nb_seq)
        nb_seq = FastaIO.nbSeq(self.tmp_multi_line_gz)
        self.assertEqual(4, nb_seq)

    def testNbSeqAndNt(self):
        nb_seq, nb_nt = FastaIO.nbSeqAndNt(self.tmp_mono_line)
        self.assertEqual(nb_seq, 4)
        self.assertEqual(nb_nt, 104)
        nb_seq, nb_nt = FastaIO.nbSeqAndNt(self.tmp_multi_line)
        self.assertEqual(nb_seq, 4)
        self.assertEqual(nb_nt, 104)
        nb_seq, nb_nt = FastaIO.nbSeqAndNt(self.tmp_multi_line_gz)
        self.assertEqual(nb_seq, 4)
        self.assertEqual(nb_nt, 104)

    def testWrite(self):
        with FastaIO(self.tmp_out, "w") as FH_out:
            for curr_rec in self.expected_rec:
                FH_out.write(curr_rec)
        self.assertTrue(FastaIO.isValid(self.tmp_out))
        self.assertTrue(filecmp.cmp(self.tmp_out, self.tmp_mono_line))

    def testIsValid(self):
        # Valid
        self.assertTrue(FastaIO.isValid(self.tmp_mono_line))
        self.assertTrue(FastaIO.isValid(self.tmp_multi_line))
        self.assertTrue(FastaIO.isValid(self.tmp_multi_line_gz))
        # Valid long file
        content = ">seq1\nATGC\n>seq2\nATGC\n>seq3\nATGC\n>seq4\nATGC\n>seq5\nATGC\n>seq6\nATGC\n>seq7\nATGC\n>seq8\nATGC\n>seq9\nATGC\n>seq10\nATGC\n>seq11\nATGC\n>seq12\nATGC"
        with open(self.tmp_out, "w") as FH_out:
            FH_out.write(content)
        self.assertTrue(FastaIO.isValid(self.tmp_out))
        # Valid empty file
        content = ""
        with open(self.tmp_out, "w") as FH_out:
            FH_out.write(content)
        self.assertTrue(FastaIO.isValid(self.tmp_out))
        # Valid empty sequence
        content = ">seq1\n"
        with open(self.tmp_out, "w") as FH_out:
            FH_out.write(content)
        self.assertTrue(FastaIO.isValid(self.tmp_out))
        # Invalid file (two headers)
        content = ">seq1\nATGC\n>seq2\n>seq3\nATGC"
        with open(self.tmp_out, "w") as FH_out:
            FH_out.write(content)
        self.assertTrue(not FastaIO.isValid(self.tmp_out))
        # Invalid file (no header at the first line)
        content = "seq1\nATGC\n>seq2\nATGC"
        with open(self.tmp_out, "w") as FH_out:
            FH_out.write(content)
        self.assertTrue(not FastaIO.isValid(self.tmp_out))
        # Invalid file (fastq)
        content = "@seq1\nATGC\n+\n####"
        with open(self.tmp_out, "w") as FH_out:
            FH_out.write(content)
        self.assertTrue(not FastaIO.isValid(self.tmp_out))

    def testIter(self):
        with FastaIO(self.tmp_mono_line) as FH_in:
            for idx, record in enumerate(FH_in):
                self.assertTrue(cmpSequences(record, self.expected_rec[idx]))
            self.assertEqual(idx + 1, 4)
        with FastaIO(self.tmp_multi_line) as FH_in:
            for idx, record in enumerate(FH_in):
                self.assertTrue(cmpSequences(record, self.expected_rec[idx]))
            self.assertEqual(idx + 1, 4)
        with FastaIO(self.tmp_multi_line_gz) as FH_in:
            for idx, record in enumerate(FH_in):
                self.assertTrue(cmpSequences(record, self.expected_rec[idx]))
            self.assertEqual(idx + 1, 4)
        with FastaIO(self.tmp_mono_line) as FH_in_mono:
            with FastaIO(self.tmp_multi_line) as FH_in_multi:
                idx = 0
                for rec_expected, rec_mono_line, rec_multi_line in zip(self.expected_rec, FH_in_mono, FH_in_multi):
                    self.assertTrue(cmpSequences(rec_mono_line, rec_expected))
                    self.assertTrue(cmpSequences(rec_multi_line, rec_expected))
                    idx += 1
                self.assertEqual(idx, 4)

    def tearDown(self):
        # Clean temporary files
        for curr_file in [self.tmp_mono_line, self.tmp_multi_line, self.tmp_multi_line_gz, self.tmp_out]:
            if os.path.exists(curr_file):
                os.remove(curr_file)


class TestFastqIO(unittest.TestCase):
    def setUp(self):
        tmp_folder = tempfile.gettempdir()
        unique_id = str(uuid.uuid1())

        # Temporary files
        self.tmp_seq = os.path.join(tmp_folder, unique_id + "_ref.fastq")
        self.tmp_seq_gz = os.path.join(tmp_folder, unique_id + "_ref.fastq.gz")
        self.tmp_out = os.path.join(tmp_folder, unique_id + "_out.fastq")

        # Expected
        self.expected_rec = [
            Sequence("seq1", "ATAGATAGCATCCCCCCNATAC", "test description", "@?>=<;:9876543210##-,+"),
            Sequence("seq2", "", None, ""),
            Sequence("seq3", "ATGAAAAAAAAAAAAANTG", "test description 2", "@?>=<;:987654321#/."),
            Sequence("seq4", "", "trimmed", "")
        ]

        # Create file
        content = """@seq1 test description
ATAGATAGCATCCCCCCNATAC
+
@?>=<;:9876543210##-,+
@seq2

+

@seq3 test description 2
ATGAAAAAAAAAAAAANTG
+
@?>=<;:987654321#/.
@seq4 trimmed

+

"""
        with open(self.tmp_seq, "w") as FH_out:
            FH_out.write(content)
        with gzip.open(self.tmp_seq_gz, "wt") as FH_out:
            FH_out.write(content)

    def testQualOffset(self):
        # Illumina 1.8 with under 59
        content = """@M70265:234:000000000-CCC3N:1:1101:21165:1697 1:N:0:13
ATGTCCTTGTGCACAATGCCCTGGCTATGCAGGTACTCCAGGCCGTCAATCAGCTGACAGAAGTACCTGCGGGCAGCACACACCCGTCCTGGGGCCGAGGCCTCCCTGCCCCTCTCAGGGGCGAATTTCGACGATCGTTGCATTAACTCGC
+
-A-A@EFF9E,C9C,,C,CEEF,,6C9E,@,,,C<EEEE,,,:B7@:,,CC,,CE,,;,,,,,<9@CE,C+++@+,@,C,C,B@>+BBFE,,,+87++++8ABA=FE,B?BFDC==,,,,+6+++@BFD,+8++>>7@D,<,@@,,@7>*>
@M70265:234:000000000-CCC3N:1:1101:14142:1764 1:N:0:13
TGTCAATCAATATCAGGACAAGCAGTGTGTCCTCACGGAAAGGAGCCTGCCCTGCCTGGCCCCCGGCCCCCGCCCCACCCTGGCCCCTGCCCCGCGCACCCACCCGTTGGCCTTGCCCCCTCGGAAACGCTTCTCCCGCACCCTTGCGAAT
+
B<CB9-CF9,F9FDC,,,C8<,C8,C,C,,<CEE9C,,+,6,<,<CBF@<@EE,CFD,,@ADCF7::C@B@+@>CC,:FFC,,4CDC<,,CED+@+>+6+8+?CC+8,+,4:>,,:,:+83++++++8+83>:,33,+3+5*68,,,**1*
@M70265:234:000000000-CCC3N:1:1101:9715:1775 1:N:0:13
TCCAGGGCTTTTGTCTTCTTCCCTTTAGATTCTCTTCTTCTGTACTGCCTGTGCTTTTGCATTCTCTACACTCATCTGTGCCACCGTTTGGAAAGCTAGTGGTTCAGAGTTCTATATATTCTCGAATTTCGCCGATCGTTTCATTAACTCT
+
-8A----8FFGG,E@E@EEF<@6CFF9,,<,;C6C,6CE@C,C,CF,@C,,;,,,<,;,;,,,6,;C,,,6;;,<<E,,,6C,C<+CBA,,,,,,6C,,:,,CC@,,,,<E@F,C,,,<EA,C,,,9?E,,,8++8>+BE+559E,,5=E,"""
        with open(self.tmp_out, "w") as FH_out:
            FH_out.write(content)
        self.assertEqual(FastqIO.qualOffset(self.tmp_out), 33)
        # Illumina 1.8
        content = """@M70265:234:000000000-CCC3N:1:1102:19767:8584 1:N:0:35
TCATGACTGATATGGTAGACAGAGCCTAAACATCCCCTTAAATTGGATTAAAAAGAAATATACCTTTGTTGTTACCTTTAAATGCAAAGTTAAAATAGGCAGAAGTCTTGCCCACATCGTTGTAGGCCTTACATTCAACCGGCGAATTTCG
+
CCCCCFGGGGGGGGGGGGGGGGGGGGGGGGGGGFGGFFGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGFFGGGGGGGGGGGGFGGGGCGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGFGGGGGGGGGGGGGGGGGEGGGGFGF
@M70265:234:000000000-CCC3N:1:1102:17014:8587 1:N:0:35
TCCATAACTTCTTGCTAAGTCCTGAGCCTGTTTTGTGTCTACTGTTCTAGAAGGCAAATCACATTTATTTCCTACTAGGACCACAGGTACATCTTCAGAGTCCTTAACTCTTTTAATTTGTTCTCTGGGAAAGAGCGAATTTCGACGATCG
+
CCCCCGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGFGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGG
@M70265:234:000000000-CCC3N:1:1102:16174:8588 1:N:0:35
CTTGAGTGAAGGACTGAGAAAATCCCTGTTCCCACTCATACAGGACTTGGGAGGTATCCACATCCTCTTCCTCAGGATTGCCTTTACCACTCTGAGAAGGAGCTGTGGTAGTGGCACCAGAATGGATTCCAGAGTCCAGGTAAGACTGCGC
+
CCCCCGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGFGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGG"""
        with open(self.tmp_out, "w") as FH_out:
            FH_out.write(content)
        self.assertEqual(FastqIO.qualOffset(self.tmp_out), 33)
        # Solexa
        content = """@SRR1296011.1 1 length=107
CGGCAAGTTAACAAAAAGAAAAATGGTGAATGATACCCGGTGCTGGCAATCTCGTTTAAACTACATGCAGGAACAGCAAAGGAAATCCGGCAAATTTGCGCAGTCAT
+SRR1296011.1 1 length=107
dddddaa]aafffc`c_ccc`cccf_^cddf_fcddd`ddWdd^a]daadf[fdcffaafcffcfcff]fcfffW^I^a^^KZdaffc_cWbc[cN[[X^]`a``ca
@SRR1296011.2 2 length=107
AAATTTGCCGGATTTCCTTTGCTGTTCCTGCATGTAGTTTAAACGAGATTGCCAGCACCGGGTATCATTCACCATTTTTCTTTTTGTTAACTTGCCGTCAGCCTTTT
+SRR1296011.2 2 length=107
gggggggaggggggggggfgfgggggggcgggggggc_geggfggggggggggaggggggggggdggffggfgggaggeegcgggeggggeffgac]dbcaggeab_
@SRR1296011.3 3 length=107
CTTTCTGTTCATGTGTATCTGCTGTCTCTTAGCCCAGACTTCCCGTGTCCTTTCCACTGGGCCTTTGGGAGGTCACAGGGTCTTGATGCTGTGGTCTTGATCTGCAG
+SRR1296011.3 3 length=107
fffdffgggggc_aggaggggfe_afffffgggggfgggggggggddgge_aWdaggggg]]cfffffedfeUeaacff_Wcfcc`bb]d__b^Zacaa[]\```_b"""
        with open(self.tmp_out, "w") as FH_out:
            FH_out.write(content)
        self.assertEqual(FastqIO.qualOffset(self.tmp_out), 64)

    def testWrite(self):
        with FastqIO(self.tmp_out, "w") as FH_out:
            for curr_rec in self.expected_rec:
                FH_out.write(curr_rec)
        self.assertTrue(FastqIO.isValid(self.tmp_out))
        self.assertTrue(filecmp.cmp(self.tmp_out, self.tmp_seq))

    def testNbSeq(self):
        nb_seq = FastqIO.nbSeq(self.tmp_seq)
        self.assertEqual(nb_seq, 4)

    def testNbSeqAndNt(self):
        nb_seq, nb_nt = FastqIO.nbSeqAndNt(self.tmp_seq)
        self.assertEqual(nb_seq, 4)
        self.assertEqual(nb_nt, 41)
        nb_seq, nb_nt = FastqIO.nbSeqAndNt(self.tmp_seq_gz)
        self.assertEqual(nb_seq, 4)
        self.assertEqual(nb_nt, 41)

    def testIsValid(self):
        # Valid
        self.assertTrue(FastqIO.isValid(self.tmp_seq))
        self.assertTrue(FastqIO.isValid(self.tmp_seq_gz))
        # Valid long file
        content = "@seq1\nATGC\n+\n####\n@seq2\nATGC\n+\n####\n@seq3\nATGC\n+\n####\n@seq4\nATGC\n+\n####\n@seq5\nATGC\n+\n####\n@seq6\nATGC\n+\n####\n@seq7\nATGC\n+\n####\n@seq8\nATGC\n+\n####\n@seq9\nATGC\n+\n####\n@seq10\nATGC\n+\n####\n@seq11\nATGC\n+\n####\n"
        with open(self.tmp_out, "w") as FH_out:
            FH_out.write(content)
        self.assertTrue(FastqIO.isValid(self.tmp_out))
        # Valid empty file
        content = ""
        with open(self.tmp_out, "w") as FH_out:
            FH_out.write(content)
        self.assertTrue(FastqIO.isValid(self.tmp_out))
        # Valid empty sequence
        content = "@seq1\n\n+\n\n"
        with open(self.tmp_out, "w") as FH_out:
            FH_out.write(content)
        self.assertTrue(FastqIO.isValid(self.tmp_out))
        # Invalid file (fasta)
        content = ">seq1\nATGC\n>seq2\nATGC"
        with open(self.tmp_out, "w") as FH_out:
            FH_out.write(content)
        self.assertTrue(not FastqIO.isValid(self.tmp_out))
        # Invalid file (not a sequence)
        content = "@seq1\nAT1GC2\n+\n######"
        with open(self.tmp_out, "w") as FH_out:
            FH_out.write(content)
        self.assertTrue(not FastqIO.isValid(self.tmp_out))

    def testIter(self):
        with FastqIO(self.tmp_seq) as FH_in:
            for idx, record in enumerate(FH_in):
                self.assertTrue(cmpSequences(record, self.expected_rec[idx]))
            self.assertEqual(idx + 1, 4)
        with FastqIO(self.tmp_seq_gz) as FH_in:
            for idx, record in enumerate(FH_in):
                self.assertTrue(cmpSequences(record, self.expected_rec[idx]))
            self.assertEqual(idx + 1, 4)
        with FastqIO(self.tmp_seq) as FH_in:
            idx = 0
            for rec_expected, rec_observed in zip(self.expected_rec, FH_in):
                self.assertTrue(cmpSequences(rec_observed, rec_expected))
                idx += 1
            self.assertEqual(idx, 4)

    def tearDown(self):
        # Clean temporary files
        for curr_file in [self.tmp_seq, self.tmp_seq_gz, self.tmp_out]:
            if os.path.exists(curr_file):
                os.remove(curr_file)


def cmpSequences(rec_1, rec_2):
    are_equal = False
    if rec_1.id == rec_2.id:
        if rec_1.description == rec_2.description:
            if rec_1.string == rec_2.string:
                if rec_1.quality == rec_2.quality:
                    are_equal = True
    return are_equal


class TestIdxFastaIO(unittest.TestCase):
    def setUp(self):
        tmp_folder = tempfile.gettempdir()
        unique_id = str(uuid.uuid1())

        # Temporary files
        self.tmp_fasta_idx = os.path.join(tmp_folder, unique_id + ".fasta.fai")
        self.tmp_fasta = os.path.join(tmp_folder, unique_id + ".fasta")

        # Expected
        self.expected_rec = {
            "one": Sequence("one", "ATGCATGCATGCATGCATGCATGCATGCATGCATGCATGCATGCATGCATGCATGCATGCATGCAT"),
            "two": Sequence("two", "ATGCATGCATGCATGCATGCATGCATGC", "another chromosome")
        }

        # Create sequence file
        content_fasta = """>one
ATGCATGCATGCATGCATGCATGCATGCAT
GCATGCATGCATGCATGCATGCATGCATGC
ATGCAT
>two another chromosome
ATGCATGCATGCAT
GCATGCATGCATGC"""
        with open(self.tmp_fasta, "w") as FH_out:
            FH_out.write(content_fasta)

        # Create index
        content_fasta_idx = """one	66	5	30	31
two	28	98	14	15"""
        with open(self.tmp_fasta_idx, "w") as FH_out:
            FH_out.write(content_fasta_idx)

    def tearDown(self):
        # Clean temporary files
        for curr_file in [self.tmp_fasta, self.tmp_fasta_idx]:
            if os.path.exists(curr_file):
                os.remove(curr_file)

    def testGet(self):
        # Without cache
        with IdxFastaIO(self.tmp_fasta) as FH:
            self.assertEqual(
                self.expected_rec["one"].string,
                FH.get("one").string
            )
            self.assertEqual(
                self.expected_rec["one"].string,
                FH.get("one").string
            )
            self.assertEqual(
                self.expected_rec["two"].string,
                FH.get("two").string
            )
            self.assertEqual(
                self.expected_rec["one"].string,
                FH.get("one").string
            )
        # With cache
        with IdxFastaIO(self.tmp_fasta, "r", True) as FH:
            self.assertEqual(
                self.expected_rec["one"].string,
                FH.get("one").string
            )
            self.assertEqual(
                self.expected_rec["one"].string,
                FH.get("one").string
            )
            self.assertEqual(
                self.expected_rec["two"].string,
                FH.get("two").string
            )
            self.assertEqual(
                self.expected_rec["one"].string,
                FH.get("one").string
            )


    def testGetSub(self):
        # From one
        one_len = len(self.expected_rec["one"].string)
        with IdxFastaIO(self.tmp_fasta) as FH:
            # Full length
            self.assertEqual(
                self.expected_rec["one"].string,
                FH.getSub("one", 1, one_len)
            )
            # Full length without end parameter
            self.assertEqual(
                self.expected_rec["one"].string,
                FH.getSub("one", 1)
            )
            # First nt
            self.assertEqual(
                self.expected_rec["one"].string[0],
                FH.getSub("one", 1, 1)
            )
            # 2 firsts nt
            self.assertEqual(
                self.expected_rec["one"].string[0:2],
                FH.getSub("one", 1, 2)
            )
            # 3 firsts nt
            self.assertEqual(
                self.expected_rec["one"].string[0:3],
                FH.getSub("one", 1, 3)
            )
            # Last nt
            self.assertEqual(
                self.expected_rec["one"].string[-1],
                FH.getSub("one", one_len, one_len)
            )
            # 2 lasts nt
            self.assertEqual(
                self.expected_rec["one"].string[-2:],
                FH.getSub("one", one_len - 1)
            )
            # 3 lasts nt
            self.assertEqual(
                self.expected_rec["one"].string[-3:],
                FH.getSub("one", one_len - 2)
            )
            # 1 nt on start of second line
            self.assertEqual(
                self.expected_rec["one"].string[30],
                FH.getSub("one", 31, 31)
            )
            # 1 nt in second line
            self.assertEqual(
                self.expected_rec["one"].string[31],
                FH.getSub("one", 32, 32)
            )
            # 11 nt in second line
            self.assertEqual(
                self.expected_rec["one"].string[31:42],
                FH.getSub("one", 32, 42)
            )
            # nt on first and second line
            self.assertEqual(
                self.expected_rec["one"].string[29:35],
                FH.getSub("one", 30, 35)
            )
            # nt on second and third line
            self.assertEqual(
                self.expected_rec["one"].string[57:63],
                FH.getSub("one", 58, 63)
            )
            # nt on first, second and third line
            self.assertEqual(
                self.expected_rec["one"].string[29:63],
                FH.getSub("one", 30, 63)
            )


class TestUtils(unittest.TestCase):
    def setUp(self):
        tmp_folder = tempfile.gettempdir()
        unique_id = str(uuid.uuid1())

        # Temporary files
        self.tmp_fasta_idx = os.path.join(tmp_folder, unique_id + ".fasta.fai")
        self.tmp_fasta = os.path.join(tmp_folder, unique_id + ".fasta")

        # Create sequence file
        content_fasta = """>one
ATGCATGCATGCATGCATGCATGCATGCAT
GCATGCATGCATGCATGCATGCATGCATGC
ATGCAT
>two another chromosome
ATGCATGCATGCAT
GCATGCATGCATGC"""
        with open(self.tmp_fasta, "w") as FH_out:
            FH_out.write(content_fasta)

        # Create index
        content_fasta_idx = """one	66	5	30	31
two	28	98	14	15"""
        with open(self.tmp_fasta_idx, "w") as FH_out:
            FH_out.write(content_fasta_idx)

    def tearDown(self):
        # Clean temporary files
        for curr_file in [self.tmp_fasta, self.tmp_fasta_idx]:
            if os.path.exists(curr_file):
                os.remove(curr_file)

    def testGetStrandedSeqFromPos(self):
        with IdxFastaIO(self.tmp_fasta) as reader:
            self.assertEqual(
                "ATGCCAT",
                getStrandedSeqFromPos("one", [5,6,7,8,20,21,34], "+", reader)
            )
            self.assertEqual(
                "ATGGCAT",
                getStrandedSeqFromPos("one", [5,6,7,8,20,21,34], "-", reader)
            )
            self.assertEqual(
                "AUGGCAU",
                getStrandedSeqFromPos("one", [5,6,7,8,20,21,34], "-", reader, RNAAlphabet)
            )


########################################################################
#
# MAIN
#
########################################################################
if __name__ == "__main__":
    unittest.main()
